# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'data_handling.database_interface' module within the ketos library
"""
import sys
import pytest
import tables
import os
import datetime as dt
import numpy as np
import pandas as pd
from io import StringIO
import ketos.data_handling.database_interface as di
import ketos.data_handling.data_handling as dh
from ketos.data_handling.parsing import parse_audio_representation
from ketos.data_handling.selection_table import use_multi_indexing 
from ketos.audio.spectrogram import MagSpectrogram, Spectrogram, CQTSpectrogram
from ketos.audio.waveform import Waveform
from ketos.audio.gammatone import GammatoneFilterBank

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')


def test_open_file_create_dir():
    """ Test if that a directory is created if it does not already exist """
    folder = os.path.join(path_to_tmp, 'new_folder')
    if os.path.isdir(folder): os.rmdir(folder) #remove folder if it already exists
    file_path = os.path.join(folder, 'db.h5')
    for mode in ['w','a']:
        assert not os.path.isdir(folder) #check that folder does not exist before attempt to open
        h5file = di.open_file(file_path, mode='w')
        h5file.close()
        assert os.path.isfile(file_path) #check that file has been created
        os.remove(file_path) #clean
        os.rmdir(folder) #clean


def test_open_non_existing_table():
    """ Test if the expected exception is raised when the table does not exist """
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp1_db.h5')
    h5file = di.open_file(fpath, 'w')
    # open non-existing table
    with pytest.raises(tables.NoSuchNodeError):
        tbl = di.open_table(h5file=h5file, table_path='/group_1/table_1')
        assert tbl == None
    # clean
    h5file.close()
    os.remove(fpath)

def test_open_existing_table():
    """ Test if the expected table is open """
    # open h5 file
    fpath = os.path.join(path_to_assets, '15x_same_spec.h5')
    h5file = di.open_file(fpath, 'r')
    # open non-existing table
    tbl = di.open_table(h5file=h5file, table_path='/train/species1')
    assert isinstance(tbl, tables.table.Table)
    assert tbl.nrows == 15
    # clean
    h5file.close()

def test_table_descr_weak():
    """ Test that we can create a table description for weakly annotated data"""
    img = np.random.random_sample((64,20))#Create a 64 x 20 image
    descr = di.table_description(img)#Create a table description for weakly labeled spectrograms of this shape
    cols = descr.columns
    keys = sorted(cols.keys())
    assert keys[0] == 'data'
    assert cols[keys[0]] == tables.Float32Col(shape=(64, 20), dflt=0.0, pos=None)
    assert keys[1] == 'filename'
    assert cols[keys[1]] == tables.StringCol(itemsize=100, shape=(), dflt=b'', pos=None)
    assert keys[2] == 'id'
    assert cols[keys[2]] == tables.UInt32Col(shape=(), dflt=0, pos=None)
    assert keys[3] == 'label'
    assert cols[keys[3]] == tables.UInt8Col(shape=(), dflt=0, pos=None)
    assert keys[4] == 'offset'
    assert cols[keys[4]] == tables.Float64Col(shape=(), dflt=0.0, pos=None)  

def test_table_descr_mult_data():
    """ Test that we can create a table description with multiple data fields"""
    img1 = np.random.random_sample((64,20))#Create a 64 x 20 image
    img2 = np.random.random_sample((32,32))#Create a 32 x 32 image
    descr = di.table_description([img1,img2])#Create a table description
    cols = descr.columns
    keys = sorted(cols.keys())
    assert keys[0] == 'data0'
    assert cols[keys[0]] == tables.Float32Col(shape=(64, 20), dflt=0.0, pos=None)
    assert keys[1] == 'data1'
    assert cols[keys[1]] == tables.Float32Col(shape=(32, 32), dflt=0.0, pos=None)

def test_table_descr_mult_data_named():
    """ Test that we can create a table description with multiple data fields
        and user specified names"""
    img1 = np.random.random_sample((64,20))#Create a 64 x 20 image
    img2 = (32,32)#32 x 32 shape
    descr, names = di.table_description([img1,img2], data_name=['a_large_img', 'a_small_img'], return_data_name=True)#Create a table description
    cols = descr.columns
    keys = sorted(cols.keys())
    assert keys[0] == 'a_large_img'
    assert cols[keys[0]] == tables.Float32Col(shape=(64, 20), dflt=0.0, pos=None)
    assert keys[1] == 'a_small_img'
    assert cols[keys[1]] == tables.Float32Col(shape=(32, 32), dflt=0.0, pos=None)
    assert names == ['a_large_img', 'a_small_img']

def test_table_descr_attrs():
    """ Test that we can create a table description with additional columns"""
    img = np.random.random_sample((64,20))
    attrs = [{'name':'a_confidence', 'shape':(), 'type':'float'},
             {'name':'a_comment', 'shape':120, 'type':'str'},
             {'name':'a_vector', 'shape':(5), 'type':'int'}]
    descr = di.table_description(img, attrs=attrs)
    cols = descr.columns
    keys = sorted(cols.keys())
    print(keys)
    assert keys[0] == 'a_comment'
    assert cols[keys[0]] == tables.StringCol(itemsize=120)
    assert keys[1] == 'a_confidence'
    assert cols[keys[1]] == tables.Float64Col(shape=(), dflt=0.0, pos=None)
    assert keys[2] == 'a_vector'
    assert cols[keys[2]] == tables.Int32Col(shape=(5,), dflt=0.0, pos=None)

def test_create_table():
    """Test if a table and its group are created"""
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp2_db.h5')
    h5file = di.open_file(fpath, 'w')
    # create table description
    descr_data = di.table_description((32,64))
    descr_annot = di.table_description_annot()
    # create data table
    _ = di.create_table(h5file=h5file, path='/group_1/', name='table_1', description=descr_data)
    group = h5file.get_node("/group_1")
    assert isinstance(group, tables.group.Group)
    table = h5file.get_node("/group_1/table_1")
    assert isinstance(table, tables.table.Table)    
    # create annotation table
    _ = di.create_table(h5file=h5file, path='/group_1/', name='table_2', description=descr_annot)
    group = h5file.get_node("/group_1")
    assert isinstance(group, tables.group.Group)
    table = h5file.get_node("/group_1/table_2")
    assert isinstance(table, tables.table.Table)    
    # clean
    h5file.close()
    os.remove(fpath)

def test_add_row_to_annot_table():
    """Test if we can add a row to the annotation table"""
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp2_db.h5')
    h5file = di.open_file(fpath, 'w')
    # create table description
    descr_annot = di.table_description_annot()
    # create annotation table
    table = di.create_table(h5file=h5file, path='/group_1/', name='table_2', description=descr_annot)
    # add a row
    row = table.row
    row['label'] = 12
    row['data_index'] = 3
    row['start'] = 0.
    row['end'] = 1.
    row.append()
    table.flush()
    assert table.nrows == 1
    # clean
    h5file.close()
    os.remove(fpath)

def test_create_table_existing():
    """Test if a table is open when it already exists"""
    # open h5 file
    fpath = os.path.join(path_to_assets, '15x_same_spec.h5')
    h5file = di.open_file(fpath, 'a')
    # create table description
    descr_annot = di.table_description_annot()
    # create table
    _ = di.create_table(h5file=h5file, path='/train/', name='species1', description=descr_annot)
    table = h5file.get_node("/train/species1")
    assert table.nrows == 15
    assert table[0]['data'].shape == (12,12)
    assert table[1]['id'] == 1
    # clean
    h5file.close()

def test_write_mag_spec(sine_audio):
    """Test if spectrograms are written and have the expected ids"""
    # create spectrogram    
    range_trans = {'name':'adjust_range', 'range':(0,1)}
    transforms = [range_trans]
    spec = MagSpectrogram.from_waveform(sine_audio, 0.5, 0.1, transforms=transforms)
    spec.filename = 'file.wav'
    spec.offset = 0.1
    # add annotation
    spec.annotate(df={'label':[1,2], 'start':[1,1.5], 'end':[2,2.5], 'freq_min':[3,3.5], 'freq_max':[4,4.5]})
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp3_db.h5')
    h5file = di.open_file(fpath, 'w')
    # Create table descriptions for storing the spectrogram data
    descr_data = di.table_description(spec)
    descr_annot = di.table_description_annot()
    # Create tables
    tbl_data = di.create_table(h5file, "/group1/", "table_data", descr_data) 
    tbl_annot = di.create_table(h5file, "/group1/", "table_annot", descr_annot) 
    # write spectrogram to table twice
    di.write(x=spec, table=tbl_data, table_annot=tbl_annot) 
    di.write(x=spec, table=tbl_data, table_annot=tbl_annot) 
    # write spectrogram to table with id
    di.write(x=spec, table=tbl_data, table_annot=tbl_annot, id=7)
    tbl_data.flush()
    tbl_annot.flush()
    # check that annotations have been properly saved
    x = tbl_annot[0]
    assert x['data_index'] == 0
    assert x['label'] == 1
    assert x['start'] == 1.
    assert x['end'] == 2.
    x = tbl_annot[1]
    assert x['data_index'] == 0
    assert x['label'] == 2
    assert x['start'] == 1.5
    assert x['end'] == 2.5
    x = tbl_annot[2]
    assert x['data_index'] == 1
    x = tbl_annot[3]
    assert x['data_index'] == 1
    x = tbl_annot[4]
    assert x['data_index'] == 2
    assert x['label'] == 1
    assert x['start'] == 1.
    assert x['end'] == 2.
    x = tbl_data[0]
    assert x['filename'].decode() == 'file.wav'
    assert x['offset'] == 0.1
    assert tbl_data[0]['id'] == 0
    assert tbl_data[1]['id'] == 1
    assert tbl_data[2]['id'] == 7
    # check that attributes have been properly saved
    assert tbl_data.attrs.audio_repres['time_res'] == 0.1
    assert tbl_data.attrs.audio_repres['freq_min'] == 0
    assert tbl_data.attrs.audio_repres['freq_res'] == 0.5 * sine_audio.rate / spec.data.shape[1]
    assert tbl_data.attrs.audio_repres['type'] == 'MagSpectrogram'
    assert tbl_data.attrs.audio_repres['window_func'] == 'hamming'
    assert tbl_data.attrs.audio_repres['transform_log'] == transforms   
    # attempt to write filename exceeding max length triggers warning
    spec.filename = "".join([str(n%10) for n in range(120)])
    with pytest.warns(UserWarning):
        di.write(x=spec, table=tbl_data, table_annot=tbl_annot)
    # clean up
    h5file.close()
    os.remove(fpath)

def test_write_cqt_spec(sine_audio):
    """Test if CQT spectrograms are written with appropriate encoding"""
    # create cqt spectrogram    
    spec = CQTSpectrogram.from_waveform(audio=sine_audio, freq_min=1, freq_max=8000, step=0.1, bins_per_oct=32)
    # add annotation
    df = {'label':[1,2], 'start':[1,1.5], 'end':[2,2.5], 'freq_min':[300,300.5], 'freq_max':[400,400.5]}
    spec.annotate(df=df)
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp12_db.h5')
    h5file = di.open_file(fpath, 'w')
    # Create table descriptions for storing the spectrogram data
    descr_data = di.table_description(spec)
    descr_annot = di.table_description_annot(freq_range=True)
    # create tables
    tbl_data  = di.create_table(h5file=h5file, path='/group_1/', name='table_data', description=descr_data)
    tbl_annot = di.create_table(h5file=h5file, path='/group_1/', name='table_annot', description=descr_annot)
    # write spectrogram to table
    di.write(x=spec, table=tbl_data, table_annot=tbl_annot) 
    tbl_data.flush()
    tbl_annot.flush()
    h5file.close()
    # re-open
    h5file = di.open_file(fpath, 'r')
    tbl_d = h5file.get_node("/group_1/table_data")
    tbl_a = h5file.get_node("/group_1/table_annot")
    # check that annotations have ben properly saved
    assert tbl_d.nrows == 1
    assert tbl_a.nrows == 2
    for i in range(2):
        assert tbl_a[i]['label']    == df['label'][i]
        assert tbl_a[i]['start']    == df['start'][i]
        assert tbl_a[i]['end']      == df['end'][i]
        assert tbl_a[i]['freq_min'] == df['freq_min'][i]
        assert tbl_a[i]['freq_max'] == df['freq_max'][i]
    # check that attributes have been properly saved
    assert tbl_d.attrs.audio_repres['time_res'] > 0.
    assert tbl_d.attrs.audio_repres['freq_min'] == 1
    assert tbl_d.attrs.audio_repres['bins_per_oct'] == 32
    assert tbl_d.attrs.audio_repres['type'] == 'CQTSpectrogram'
    assert tbl_d.attrs.audio_repres['window_func'] == 'hann'
    # clean
    h5file.close()
    os.remove(fpath)

def test_write_multiple_audio_objects(sine_audio):
    """Test if multiple audio objects can be written to an HDF5 table"""
    # create spectrogram    
    spec = MagSpectrogram.from_waveform(sine_audio, 0.5, 0.1)
    spec.filename = 'file.wav'
    spec.offset = 0.1
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp33_db.h5')
    h5file = di.open_file(fpath, 'w')
    # Create table descriptions for storing the spectrogram and waveform data
    descr, data_name = di.table_description([spec, sine_audio], data_name=['spec','waveform'], return_data_name=True)
    # Create tables
    tbl = di.create_table(h5file, "/group1/", "table_data", description=descr, data_name=data_name) 
    # write spectrogram and waveforms to table twice
    di.write(x=[spec, sine_audio], table=tbl) 
    di.write(x=[spec, sine_audio], table=tbl) 
    # write spectrogram and waveform to table with id
    di.write(x=[spec, sine_audio], table=tbl, id=7)
    tbl.flush()
    # check spectrogram 
    x = tbl[0]
    assert x['filename'].decode() == 'file.wav'
    assert x['offset'] == 0.1
    assert tbl[0]['id'] == 0
    assert tbl[1]['id'] == 1
    assert tbl[2]['id'] == 7
    # check that spectrogram representation has been properly saved
    assert tbl.attrs.audio_repres[0]['time_res'] == 0.1
    assert tbl.attrs.audio_repres[0]['freq_min'] == 0
    assert tbl.attrs.audio_repres[0]['freq_res'] == 0.5 * sine_audio.rate / spec.data.shape[1]
    assert tbl.attrs.audio_repres[0]['type'] == 'MagSpectrogram'
    assert tbl.attrs.audio_repres[0]['window_func'] == 'hamming'
    # check that waveform representation has been properly saved
    assert tbl.attrs.audio_repres[1]['rate'] == sine_audio.rate
    assert tbl.attrs.audio_repres[1]['type'] == 'Waveform'
    # test load audio

    assert len(tbl) == 3
    assert len(tbl[0]) == 6
    np.testing.assert_array_almost_equal(tbl[0]['spec'], spec.data.data, decimal=3)
    np.testing.assert_array_almost_equal(tbl[0]['waveform'], sine_audio.data)
    # clean up
    h5file.close()
    os.remove(fpath)

def test_write_spectrogram_and_numpy_array(sine_audio):
    """Test if multiple audio objects can be written to an HDF5 table where one 
        of the objects is a numpy array"""
    # create spectrogram    
    spec = MagSpectrogram.from_waveform(sine_audio, 0.5, 0.1)
    spec.filename = 'file.wav'
    spec.offset = 0.1
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp33_db.h5')
    h5file = di.open_file(fpath, 'w')
    # Create table descriptions for storing the spectrogram and waveform data
    descr, data_name = di.table_description([spec, (6,12)], data_name=['spec','features'], return_data_name=True)
    # Create tables
    tbl = di.create_table(h5file, "/group1/", "table_data", description=descr, data_name=data_name) 
    # write spectrogram and numpy arrays to table twice
    di.write(x=[spec, np.ones(shape=(6,12))], table=tbl) 
    di.write(x=[spec, np.ones(shape=(6,12))], table=tbl) 
    tbl.flush()
    # check spectrogram 
    x = tbl[0]
    assert x['filename'].decode() == 'file.wav'
    assert x['offset'] == 0.1
    assert tbl[0]['id'] == 0
    assert tbl[1]['id'] == 1
    # check that spectrogram representation has been properly saved
    assert tbl.attrs.audio_repres[0]['time_res'] == 0.1
    assert tbl.attrs.audio_repres[0]['freq_min'] == 0
    assert tbl.attrs.audio_repres[0]['freq_res'] == 0.5 * sine_audio.rate / spec.data.shape[1]
    assert tbl.attrs.audio_repres[0]['type'] == 'MagSpectrogram'
    assert tbl.attrs.audio_repres[0]['window_func'] == 'hamming'
    # check that numpy array has been properly saved
    assert tbl.attrs.audio_repres[1]['type'] == 'numpy.ndarray'
    # test load audio
    assert len(tbl) == 2
    assert len(tbl[0]) == 6
    np.testing.assert_array_almost_equal(tbl[0]['spec'], spec.data.data, decimal=3)
    np.testing.assert_array_almost_equal(tbl[0]['features'], np.ones(shape=(6,12)))
    # clean up
    h5file.close()
    os.remove(fpath)

def test_write_spectrogram_with_extra_columns(sine_audio):
    """Test if an audio objects with additional columns can be written to 
        an HDF5 table"""
    # create spectrogram    
    extra_columns = {'confidence':0.3, 'comment':'funny', 'vector':np.array([1,2,3], dtype=np.int32)}
    spec = MagSpectrogram.from_waveform(sine_audio, 0.5, 0.1, **extra_columns)
    spec.filename = 'file.wav'
    spec.offset = 0.1
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp34_db.h5')
    h5file = di.open_file(fpath, 'w')
    # additiona columns that we want to save
    attrs = [{'name':'confidence', 'shape':(), 'type':'float'},
             {'name':'comment', 'shape':120, 'type':'str'},
             {'name':'vector', 'shape':(3), 'type':'int'}]
    # Create table descriptions for storing the spectrogram and waveform data
    descr, data_name = di.table_description([spec, (6,12)], data_name=['spec','features'], return_data_name=True, attrs=attrs)
    # Create tables
    tbl = di.create_table(h5file, "/group1/", "table_data", description=descr, data_name=data_name) 
    # write spectrogram and numpy arrays to table twice
    di.write(x=[spec, np.ones(shape=(6,12))], table=tbl) 
    di.write(x=[spec, np.ones(shape=(6,12))], table=tbl) 
    tbl.flush()
    # check spectrogram 
    x = tbl[0]
    assert x['filename'].decode() == 'file.wav'
    assert x['offset'] == 0.1
    assert x['confidence'] == 0.3
    assert x['comment'].decode() == 'funny'
    assert np.all(x['vector'] == np.array([1,2,3], dtype=int))
    assert tbl[0]['id'] == 0
    assert tbl[1]['id'] == 1
    # check that spectrogram representation has been properly saved
    assert tbl.attrs.audio_repres[0]['time_res'] == 0.1
    assert tbl.attrs.audio_repres[0]['freq_min'] == 0
    assert tbl.attrs.audio_repres[0]['freq_res'] == 0.5 * sine_audio.rate / spec.data.shape[1]
    assert tbl.attrs.audio_repres[0]['type'] == 'MagSpectrogram'
    assert tbl.attrs.audio_repres[0]['window_func'] == 'hamming'
    # check that numpy array has been properly saved
    assert tbl.attrs.audio_repres[1]['type'] == 'numpy.ndarray'
    # test load audio
    # selected_obj = di.load_audio(table=tbl)
    assert len(tbl) == 2
    assert len(tbl[0]) == 9
    np.testing.assert_array_almost_equal(tbl[0]['spec'], spec.data.data, decimal=3)
    np.testing.assert_array_almost_equal(tbl[0]['features'], np.ones(shape=(6,12)))
    expected = {'filename':'file.wav', 'offset':0.1, 'label':0}
    expected.update(extra_columns)

    assert expected['filename'] == tbl[0]['filename'].decode('latin-1')
    assert expected['offset'] == tbl[0]['offset']
    assert expected['confidence'] == tbl[0]['confidence']
    np.testing.assert_array_almost_equal(expected['vector'], tbl[0]['vector'])
    assert expected['comment'] == tbl[0]['comment'].decode('latin-1')
    # clean up
    h5file.close()
    os.remove(fpath)

def test_filter_by_label(sine_audio):
    """ Test if filter_by_label works when providing an int or list of ints as the label argument"""
    # create spectrogram  
    spec1 = MagSpectrogram.from_waveform(sine_audio, window=0.2, step=0.02)
    spec1.annotate(label=1, start=1.0, end=1.4, freq_min=50, freq_max=300)
    spec2 = MagSpectrogram.from_waveform(sine_audio, window=0.2, step=0.02)
    spec2.annotate(label=2, start=1.0, end=1.4, freq_min=50, freq_max=300)
    spec3 = MagSpectrogram.from_waveform(sine_audio, window=0.2, step=0.02)
    spec3.annotate(df={'label':[2,3], 'start':[1.0,2.0], 'end':[1.4,2.4], 'freq_min':[50,60], 'freq_max':[300,200]})
    # open h5 file
    fpath = os.path.join(path_to_tmp, 'tmp8_db.h5')
    h5file = di.open_file(fpath, 'w')
    # Create table descriptions for storing the spectrogram data
    descr_data = di.table_description(spec1)
    descr_annot = di.table_description_annot(freq_range=True)
    # create tables
    tbl_data  = di.create_table(h5file=h5file, path='/group_1/', name='table_data', description=descr_data)
    tbl_annot = di.create_table(h5file=h5file, path='/group_1/', name='table_annot', description=descr_annot)
    # write spectrogram to table
    di.write(x=spec1, table=tbl_data, table_annot=tbl_annot) 
    di.write(x=spec1, table=tbl_data, table_annot=tbl_annot) 
    di.write(x=spec2, table=tbl_data, table_annot=tbl_annot) 
    di.write(x=spec2, table=tbl_data, table_annot=tbl_annot)
    di.write(x=spec3, table=tbl_data, table_annot=tbl_annot) 
    tbl_data.flush()
    tbl_annot.flush()
    # select spectrograms containing the label 1
    rows = di.filter_by_label(table=tbl_annot, label=1)
    assert len(rows) == 2
    assert rows == [0,1]
    # select spectrograms containing the label 2
    rows = di.filter_by_label(table=tbl_annot, label=[2])
    assert len(rows) == 3
    assert rows == [2,3,4]
    # select spectrograms containing the labels 1 or 3
    rows = di.filter_by_label(table=tbl_annot, label=[1,3])
    assert len(rows) == 3
    assert rows == [0,1,4]
    h5file.close()
    os.remove(fpath)

def test_filter_by_label_raises_exception(sine_audio):
    """ Test if filter_by_label raises expected exception when the the label argument is of the wrong type"""
    # open h5 file
    fpath = os.path.join(path_to_assets, '15x_same_spec.h5')
    h5file = di.open_file(fpath, 'r')
    tbl = di.open_table(h5file,"/train/species1")
    
    with pytest.raises(TypeError):
        di.filter_by_label(table=tbl, label='a')
    with pytest.raises(TypeError):
        di.filter_by_label(table=tbl, label=['a','b'])
    with pytest.raises(TypeError):        
        di.filter_by_label(table=tbl, label='1')
    with pytest.raises(TypeError):    
        di.filter_by_label(table=tbl, label='1,2')
    with pytest.raises(TypeError):    
        di.filter_by_label(table=tbl, label=b'1')
    with pytest.raises(TypeError):    
        di.filter_by_label(table=tbl, label=1.0)
    with pytest.raises(TypeError):    
        di.filter_by_label(table=tbl, label= (1,2))
    with pytest.raises(TypeError):    
        di.filter_by_label(table=tbl, label=[1.0,2])
   
    h5file.close()

def test_load_audio_no_index_list():
    """Test if load audio loads the entire table if index_list is None""" 
    fpath = os.path.join(path_to_assets, '11x_same_spec.h5')
    h5file = di.open_file(fpath, 'r')
    tbl_data = di.open_table(h5file,"/group_1/table_data")
    tbl_annot = di.open_table(h5file,"/group_1/table_annot")    
    selected_specs = di.load_audio(table=tbl_data, table_annot=tbl_annot)
    assert len(selected_specs) == tbl_data.nrows
    is_spec = [isinstance(item, Spectrogram) for item in selected_specs]
    assert all(is_spec)    
    h5file.close()

def test_load_audio_with_index_list():
    """Test if load_audio loads the spectrograms specified by index_list""" 
    fpath = os.path.join(path_to_assets, '11x_same_spec.h5')
    h5file = di.open_file(fpath, 'r')
    tbl_data = di.open_table(h5file,"/group_1/table_data")
    tbl_annot = di.open_table(h5file,"/group_1/table_annot")    
    selected_specs = di.load_audio(table=tbl_data, table_annot=tbl_annot, indices=[0,3,10])
    assert len(selected_specs) == 3
    is_spec = [isinstance(item, Spectrogram) for item in selected_specs]
    assert all(is_spec)
    h5file.close()

def test_load_audio_also_loads_annotations():
    """Test if the spectrograms returned by load_audio have annotations""" 
    fpath = os.path.join(path_to_assets, '11x_same_spec.h5')
    h5file = di.open_file(fpath, 'r')
    tbl_data = di.open_table(h5file,"/group_1/table_data")
    tbl_annot = di.open_table(h5file,"/group_1/table_annot")    
    specs = di.load_audio(table=tbl_data, table_annot=tbl_annot, indices=[0,3,10])
    # check annotations for 1st spec
    d = '''label  start  end  freq_min  freq_max
0      2    1.0  1.4      50.0     300.0
1      3    2.0  2.4      60.0     200.0'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0])
    res = specs[0].get_annotations()[ans.columns.values].astype({'freq_min': 'float64', 'freq_max': 'float64'})
    pd.testing.assert_frame_equal(ans, res)
    assert specs[0].filename == 'sine_wave'
    h5file.close()

def test_init_audio_writer():
    out = os.path.join(path_to_tmp, 'db4.h5')
    di.AudioWriter(output_file=out)

def test_audio_writer_can_write_one_spec(sine_wave_file):
    range_trans = {'name':'adjust_range', 'range':(0,1)}
    transforms = [range_trans]
    noise_trans = {'name':'add_gaussian_noise', 'sigma':2.0}
    wf_transforms = [noise_trans]
    sine_audio = Waveform.from_wav(sine_wave_file, transforms=wf_transforms)
    out = os.path.join(path_to_tmp, 'db5.h5')
    writer = di.AudioWriter(output_file=out)
    spec = MagSpectrogram.from_waveform(sine_audio, 0.5, 0.1, transforms=transforms)
    writer.write(spec)
    writer.close()
    fname = os.path.join(path_to_tmp, 'db5.h5')
    fil = di.open_file(fname, 'r')
    assert '/audio' in fil
    table = di.open_table(fil, '/audio')
    assert len(table) == 1
    assert table.attrs.audio_repres['transform_log'] == transforms
    assert table.attrs.audio_repres['waveform_transform_log'] == wf_transforms
    fil.close()

def test_audio_writer_can_write_one_gammatone_filter_bank(sine_audio):
    out = os.path.join(path_to_tmp, 'db14.h5')
    writer = di.AudioWriter(output_file=out)
    gfb = GammatoneFilterBank.from_waveform(sine_audio, num_chan=20, freq_min=10)
    writer.write(gfb)
    writer.close()
    fname = os.path.join(path_to_tmp, 'db14.h5')
    fil = di.open_file(fname, 'r')
    assert '/audio' in fil
    table = di.open_table(fil, '/audio')
    assert len(table) == 1
    assert table.attrs.audio_repres['weight_func'] == True
    assert len(table.attrs.audio_repres['freqs']) == 20
    assert np.min(table.attrs.audio_repres['freqs']) == pytest.approx(10., abs=1e-6)
    assert table[0][0].shape[1] == 20
    fil.close()

def test_audio_writer_can_write_two_specs_to_same_node(sine_audio):
    out = os.path.join(path_to_tmp, 'db6.h5')
    writer = di.AudioWriter(output_file=out)
    spec = MagSpectrogram.from_waveform(sine_audio, 0.5, 0.1)
    writer.write(spec)
    writer.write(spec)
    writer.close()
    fname = os.path.join(path_to_tmp, 'db6.h5')
    fil = di.open_file(fname, 'r')
    assert '/audio' in fil
    table = di.open_table(fil, '/audio')
    assert len(table) == 2
    fil.close()

def test_audio_writer_can_write_several_specs_to_different_nodes(sine_audio):
    out = os.path.join(path_to_tmp, 'db7.h5')
    writer = di.AudioWriter(output_file=out)
    spec = MagSpectrogram.from_waveform(sine_audio, 0.5, 0.1)
    writer.write(spec, path='/first', name='test')
    writer.write(spec, path='/first', name='test')
    writer.write(spec, path='/second', name='temp')
    writer.write(spec, path='/second', name='temp')
    writer.write(spec, path='/second', name='temp')
    writer.close()
    fname = os.path.join(path_to_tmp, 'db7.h5')
    fil = di.open_file(fname, 'r')
    assert '/first/test' in fil
    assert '/second/temp' in fil
    node1 = di.open_table(fil, '/first/test')
    node2 = di.open_table(fil, '/second/temp')
    assert len(node1) == 2
    assert len(node2) == 3
    fil.close()

def test_audio_writer_splits_into_several_files_when_max_size_is_reached(sine_audio):
    out = os.path.join(path_to_tmp, 'db8.h5')
    writer = di.AudioWriter(output_file=out, max_size=2E6) # max size: 1 Mbyte
    spec = MagSpectrogram.from_waveform(sine_audio, 0.5, 0.1)
    writer.write(spec)
    writer.write(spec)
    writer.write(spec)
    writer.close()

    fname = os.path.join(path_to_tmp, 'db8_000.h5')
    fil = di.open_file(fname, 'r')
    assert '/audio' in fil
    table = di.open_table(fil, '/audio')
    assert len(table) == 2
    fil.close()

    fname = os.path.join(path_to_tmp, 'db8_001.h5')
    fil = di.open_file(fname, 'r')
    assert '/audio' in fil
    table = di.open_table(fil, '/audio')
    assert len(table) == 1
    fil.close()

def test_audio_writer_change_directory(sine_audio):
    out = os.path.join(path_to_tmp, 'db9.h5')
    writer = di.AudioWriter(output_file=out)
    spec = MagSpectrogram.from_waveform(sine_audio, 0.5, 0.1)
    writer.set_table('/home/','fish')
    writer.write(spec)
    writer.write(spec)
    writer.write(spec)
    writer.set_table('/home','whale')
    writer.write(spec)
    writer.write(spec)
    writer.close()
    fname = os.path.join(path_to_tmp, 'db9.h5')
    fil = di.open_file(fname, 'r')
    assert '/home/fish' in fil
    assert '/home/whale' in fil
    table = di.open_table(fil, '/home/fish')
    assert len(table) == 3
    table = di.open_table(fil, '/home/whale')
    assert len(table) == 2
    fil.close()

def test_two_audio_writers_simultaneously(sine_audio):
    # init two spec writers
    out1 = os.path.join(path_to_tmp, 'db10.h5')
    writer1 = di.AudioWriter(output_file=out1)
    out2 = os.path.join(path_to_tmp, 'db11.h5')
    writer2 = di.AudioWriter(output_file=out2)
    # create spec
    spec = MagSpectrogram.from_waveform(sine_audio, 0.5, 0.1)
    # write 
    writer1.set_table('/home/','fish')
    writer1.write(spec)
    writer1.write(spec)
    writer1.write(spec)
    writer2.set_table('/home/','whale')
    writer2.write(spec)
    writer2.write(spec)
    # close
    writer1.close()
    writer2.close()
    # check file 1
    fil1 = di.open_file(out1, 'r')
    assert '/home/fish' in fil1
    table = di.open_table(fil1, '/home/fish')
    assert len(table) == 3
    fil1.close()
    # check file 2
    fil2 = di.open_file(out2, 'r')
    assert '/home/whale' in fil2
    table = di.open_table(fil2, '/home/whale')
    assert len(table) == 2
    fil2.close()

def test_create_database_with_single_wav_file(sine_wave_file):
    data_dir = os.path.dirname(sine_wave_file)
    out = os.path.join(path_to_tmp, 'db12.h5')
    rep = parse_audio_representation({'type': 'MagSpectrogram', 'window':0.5, 'step':0.1})
    sel = pd.DataFrame({'filename':['sine_wave.wav','sine_wave.wav'], 'start':[0.1,0.2], 'end':[2.0,2.1], 'label':[1,2]})
    sel = use_multi_indexing(sel, 'sel_id')
    di.create_database(out, data_dir=data_dir, selections=sel, audio_repres=rep, verbose=False, progress_bar=False)
    # check database contents
    fil = di.open_file(out, 'r')
    assert '/tmp/data' in fil
    table = di.open_table(fil, '/tmp/data')
    assert len(table) == 2
    assert np.all(di.open_table(fil, "/tmp/data").attrs.unique_labels == [1,2])
    fil.close()
    os.remove(out)

def test_create_database_specify_labels(sine_wave_file):
    data_dir = os.path.dirname(sine_wave_file)
    out = os.path.join(path_to_tmp, 'db12.h5')
    rep = parse_audio_representation({'type': 'MagSpectrogram', 'window':0.5, 'step':0.1})
    sel = pd.DataFrame({'filename':['sine_wave.wav','sine_wave.wav'], 'start':[0.1,0.2], 'end':[2.0,2.1], 'label':[1,2]})
    sel = use_multi_indexing(sel, 'sel_id')
    di.create_database(out, data_dir=data_dir, selections=sel, audio_repres=rep, verbose=False, progress_bar=False, unique_labels=[3,4,5])
    # check database contents
    fil = di.open_file(out, 'r')
    assert np.all(di.open_table(fil, "/tmp/data").attrs.unique_labels == [3,4,5])
    fil.close()
    os.remove(out)

def test_create_database_ids(sine_wave_file):
    data_dir = os.path.dirname(sine_wave_file)
    out = os.path.join(path_to_tmp, 'db13.h5')
    rep = parse_audio_representation({'type': 'MagSpectrogram', 'window':0.5, 'step':0.1})
    sel = pd.DataFrame({'filename':['sine_wave.wav','sine_wave.wav', 'sine_wave.wav','sine_wave.wav'], 'start':[0.1, 0.2, 0.1, 0.2], 'end':[2.0, 2.1, 2.0, 2.1], 'label':[1, 2, 1, 2]})
    sel = use_multi_indexing(sel, 'sel_id')
    di.create_database(out, data_dir=data_dir, dataset_name='test', selections=sel, audio_repres=rep, verbose=False, progress_bar=False)
    # check database contents
    db = di.open_file(out, 'r')
    data_table = db.get_node("/test/data")
    np.testing.assert_array_equal(data_table[:]['id'],[0,1,2,3])
    assert np.all(di.open_table(db, "/test/data").attrs.unique_labels == [1,2])
    db.close()
    os.remove(out)

def test_create_database_with_single_wav_file_mult_repres(sine_wave_file):
    data_dir = os.path.dirname(sine_wave_file)
    out = os.path.join(path_to_tmp, 'db14.h5')
    rep1 = {'type': Waveform}
    rep2 = {'type': MagSpectrogram, 'window':0.5, 'step':0.1}
    sel = pd.DataFrame({'filename':['sine_wave.wav','sine_wave.wav'], 'start':[0.1,0.2], 'end':[2.0,2.1], 'label':[1,2]})
    sel = use_multi_indexing(sel, 'sel_id')
    di.create_database(out, data_dir=data_dir, selections=sel, audio_repres={'wf':rep1,'spec':rep2}, verbose=False, progress_bar=False)
    # check database contents
    fil = di.open_file(out, 'r')
    assert '/tmp/data' in fil
    table = di.open_table(fil, '/tmp/data')
    assert len(table) == 2

    spec = MagSpectrogram.from_wav(sine_wave_file, window=0.5, step=0.1, duration=1.9, offset=0.1)
    waveform = Waveform.from_wav(sine_wave_file, duration=1.9, offset=0.1)
    np.testing.assert_array_almost_equal(table[0]['wf'], waveform.data)
    np.testing.assert_array_almost_equal(table[0]['spec'], spec.data.data, decimal=3)
    assert fil.root.tmp.data.attrs.data_name == ['wf','spec']
    fil.close()
    os.remove(out)
    # check that it also works if the audio representations are specified as a list
    di.create_database(out, data_dir=data_dir, selections=sel, audio_repres=[rep1,rep2], verbose=False, progress_bar=False)
    # check database contents
    fil = di.open_file(out, 'r')
    assert '/tmp/data' in fil
    table = di.open_table(fil, '/tmp/data')
    assert len(table) == 2
    spec = MagSpectrogram.from_wav(sine_wave_file, window=0.5, step=0.1, duration=1.9, offset=0.1)
    waveform = Waveform.from_wav(sine_wave_file, duration=1.9, offset=0.1)
    np.testing.assert_array_almost_equal(table[0][0], waveform.data)
    np.testing.assert_array_almost_equal(table[0][1], spec.data.data, decimal=3)
    assert fil.root.tmp.data.attrs.data_name == ['data0','data1']
    fil.close()
    os.remove(out)

def test_create_database_with_extra_columns(sine_wave_file):
    """ Create a database including extra columns from the selection table"""
    data_dir = os.path.dirname(sine_wave_file)
    out = os.path.join(path_to_tmp, 'db16.h5')
    rep = {'type': Waveform}
    t = [dt.datetime(2002,2,23,14,20,30), dt.datetime(2012,2,23,14,20,30)]
    sel = pd.DataFrame({'filename':['sine_wave.wav','sine_wave.wav'], 'start':[0.1,0.2], 'end':[2.0,2.1], 'comment':['big','bigger'], 't':t})
    sel = use_multi_indexing(sel, 'sel_id')
    di.create_database(out, data_dir=data_dir, selections=sel, audio_repres=rep, verbose=False, progress_bar=False, include_attrs=True)
    # check database contents
    fil = di.open_file(out, 'r')
    assert '/tmp/data' in fil
    table = di.open_table(fil, '/tmp/data')
    waveform = Waveform.from_wav(sine_wave_file, duration=1.9, offset=0.1)
    np.testing.assert_array_almost_equal(table[0]['data'], waveform.data)
    assert table[0]['comment'].decode('latin-1') == 'big'
    assert table[1]['comment'].decode('latin-1') == 'bigger'
    assert table[0]['t'].decode('latin-1') == t[0].strftime("%Y%m%d_%H:%M:%S")  #currently, we are not converting string to datetime object when loading
    assert table[1]['t'].decode('latin-1') == t[1].strftime("%Y%m%d_%H:%M:%S")  #currently, we are not converting string to datetime object when loading

    fil.close()
    os.remove(out)

def test_create_database_with_indices(sine_wave_file):
    """ Create a database with indices for some columns"""
    data_dir = os.path.dirname(sine_wave_file)
    out = os.path.join(path_to_tmp, 'db17.h5')
    rep = {'type': Waveform}
    sel = pd.DataFrame({'filename':['sine_wave.wav','sine_wave.wav'], 'start':[0.1,0.2], 'end':[2.0,2.1], 'extra_id':[13,14]})
    sel = use_multi_indexing(sel, 'sel_id')
    di.create_database(out, data_dir=data_dir, selections=sel, audio_repres=rep, verbose=False, 
        progress_bar=False, include_attrs=True, index_cols=["filename", "extra_id"])
    # check database contents
    fil = di.open_file(out, 'r')
    assert '/tmp/data' in fil
    table = di.open_table(fil, '/tmp/data')

    waveform = Waveform.from_wav(sine_wave_file, duration=1.9, offset=0.1)
    np.testing.assert_array_almost_equal(table[0]['data'], waveform.data)
    assert table[0]['extra_id'] == 13
    assert table[1]['extra_id'] == 14
    fil.close()
    os.remove(out)

def test_create_database_with_exception_handling(sine_wave_file):
    """ Check if database can be created if file does not exist """
    data_dir = os.path.dirname(sine_wave_file)
    out = os.path.join(path_to_tmp, 'db18.h5')
    rep = {'type': MagSpectrogram, 'window':0.5, 'step':0.1}
    sel = pd.DataFrame({'filename':['sine_wave.wav', 'unknown_file_1.wav', 'sine_wave.wav', 'unknown_file_2.wav'], 
                        'start':[0.1, 0.2, 0.3, 0.4], 
                        'end':[2.0, 2.1, 2.2, 2.3], 
                        'label':[1, 2, 1, 1]})
    sel = use_multi_indexing(sel, 'sel_id')
    di.create_database(out, data_dir=data_dir, selections=sel, audio_repres=rep, verbose=True, progress_bar=False)
    # check database contents
    db = di.open_file(out, 'r')
    assert '/tmp/data' in db
    table = di.open_table(db, '/tmp/data')
    assert len(table) == 2
    db.close()
    os.remove(out)
    
def test_create_database_check_file_duration(sine_wave_file):
    """ Check if database can be created if selections are outside file """
    data_dir = os.path.dirname(sine_wave_file)
    out = os.path.join(path_to_tmp, 'db19.h5')
    rep = {'type': MagSpectrogram, 'window':0.5, 'step':0.1}
    sel = pd.DataFrame({'filename':['sine_wave.wav', 'sine_wave.wav', 'sine_wave.wav', 'sine_wave.wav', 'sine_wave.wav'], 
                        'start':[2.6, 0.1, -11.0, 4.5, -0.7],
                        'end':[3.6, 1.1, -8.0, 5.5, 0.3],
                        'label':[1, 1, 2, 1, 2]})
    sel = use_multi_indexing(sel, 'sel_id')
    di.create_database(out, data_dir=data_dir, selections=sel, audio_repres=rep, verbose=True, progress_bar=False)
    # check database contents
    db = di.open_file(out, 'r')
    assert '/tmp/data' in db
    table = di.open_table(db, '/tmp/data')
    assert len(table) == 3
    db.close()
    os.remove(out)
    


# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'audio.base_audio' module within the ketos library
"""
import pytest
import ketos.audio.base_audio as aba
import numpy as np


def test_init(base_audio_1d):
    """Test if the BaseAudio object has expected attribute values"""
    o, d = base_audio_1d
    assert o.ndim == 1
    assert o.filename == 'x'
    assert o.offset == 2.
    assert o.label == 13

def test_data(base_audio_1d):
    """Test that the data attribute has expect contents"""
    o, d = base_audio_1d
    assert np.all(o.data == d)

def test_get_data(base_audio_1d):
    """Test that the get_data method returns expected result"""
    o, d = base_audio_1d
    assert np.all(o.get_data() == d)

def test_crop(base_audio_time_1d):
    """Test if a cropped BaseAudio object has the expected content and length"""
    o, d = base_audio_time_1d
    oc = o.crop(start=0.2, end=3.8)
    assert oc.duration() == 3.6
    assert np.all(oc.data == d[200:3800])

def test_annotations_returns_none(base_audio_1d):
    """Test that no annotations are returned when none are provided"""
    o, d = base_audio_1d
    assert o.get_annotations() == None

def test_time_res(base_audio_time_1d):
    """Test if time resolution is correct"""
    o, d = base_audio_time_1d
    assert o.time_res() == 0.001

def test_deep_copy(base_audio_time_1d):
    """Test that changes to copy to not affect original instance"""
    o, d = base_audio_time_1d
    oc = o.deepcopy()
    oc.filename = 'yy'
    oc.time_ax.label = 'new axis'
    assert o.filename == 'x'
    assert o.time_ax.label == 'Time (s)'

def test_normalize(base_audio_1d):
    """Test that object is properly normalized"""
    N = 10000
    d = np.arange(N)
    o = aba.BaseAudio(data=d, filename='x', offset=2., duration=10., label=13)
    o.normalize()
    assert np.all(np.isclose(np.mean(o.data, axis=0), 0, atol=1e-10))
    assert np.all(np.isclose(np.std(o.data, axis=0), 1, atol=1e-10))
    o.normalize(mean=0.2, std=7.02)
    assert np.all(np.isclose(np.mean(o.data, axis=0), 0.2, atol=1e-10))
    assert np.all(np.isclose(np.std(o.data, axis=0), 7.02, atol=1e-10))

def test_adjust_range(base_audio_1d):
    """Test that object isproperly transformed to the specified range"""
    N = 10000
    d = np.arange(N)
    o = aba.BaseAudio(data=d, filename='x', offset=2., duration=10., label=13)
    o.adjust_range()
    assert np.all(np.min(o.data, axis=0) == 0)
    assert np.all(np.max(o.data, axis=0) == 1)
    o.adjust_range(range=(-1,3.5))
    assert np.all(np.min(o.data, axis=0) == -1)
    assert np.all(np.max(o.data, axis=0) == 3.5)

def test_segment(base_audio_time_1d):
    """Test segment method on 1d object"""
    o, d = base_audio_time_1d
    ss = o.segment(window=2, step=1) #integer number of steps
    assert len(ss) == 9
    for i,s in enumerate(ss):
        assert s.ndim == o.ndim
        assert s.data.shape == (2000,)
        assert s.label == 13
        assert s.offset == o.offset + 1.0 * i
    ss = o.segment(window=2, step=1.1) #non-integer number of steps
    assert len(ss) == 9
    for i,s in enumerate(ss):
        assert s.ndim == o.ndim
        assert s.data.shape == (2000,)
        if i==len(ss)-1: assert np.all(s.data[1200:] == 0) #last frame was padded with zeros
        assert s.offset == o.offset + 1.1 * i

def test_annotate(base_audio_1d):
    """Test that we can add annotations"""
    o, d = base_audio_1d
    o.annotate(label=1, start=0.2, end=1.3) 
    o.annotate(label=2, start=1.8, end=2.2) 
    assert o.annot.num_annotations() == 2

def test_label_array(base_audio_time_1d):
    """Check that method label_array returns expected array"""
    o, d = base_audio_time_1d
    o.annotate(label=1, start=0.2, end=1.3) 
    o.annotate(label=1, start=1.8, end=2.2) 
    res = o.label_array(1)
    ans = np.concatenate([np.zeros(200),np.ones(1100),np.zeros(500),np.ones(400),np.zeros(7800)]) 
    assert np.all(res == ans)

def test_segment_with_annotations(base_audio_time_1d):
    """Test segment method on 1d object with annotations"""
    o, d = base_audio_time_1d
    o.annotate(label=1, start=0.2, end=1.3) #fully contained in first segment, partially contained in second
    o.annotate(label=2, start=1.8, end=2.2) #partially contained in first, second and third segment
    ss = o.segment(window=2, step=1) 
    df0 = ss[0].get_annotations() #annotations for 1st segment
    assert len(df0) == 2
    assert np.all(df0['start'].values == [0.2, 1.8])
    assert np.all(df0['end'].values == [1.3, 2.0])
    df1 = ss[1].get_annotations() #annotations for 2nd segment
    assert len(df1) == 2
    assert np.all(df1['start'].values == [0.0, 0.8])
    assert np.all(np.abs(df1['end'].values - [0.3, 1.2]) < 1e-9)

def test_get_attrs():
    """Test that base audio returns attrs"""
    N = 10000
    d = np.arange(N)
    o = aba.BaseAudio(data=d, filename='x', offset=2., duration=10., label=13, confidence=0.5)
    x = o.get_instance_attrs()
    assert x == {'filename':'x', 'offset':2., 'duration':10., 'label':13, 'confidence':0.5}
    o = aba.BaseAudioTime(data=d, time_res=0.001, filename='x', offset=2., label=13, confidence=0.5)
    x = o.get_instance_attrs()
    assert x == {'filename':'x', 'offset':2., 'label':13, 'confidence':0.5}



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" 'neural_networks.dev_utils.export' module within the ketos library

    This module contains utilities for saving ketos models in various formats.
"""

import ketos.data_handling.parsing as pa
from tensorflow.saved_model import save as save_pb
# from ketos.audio.audio_loader import audio_repres_dict
from ketos.utils import ensure_dir
from zipfile import ZipFile
import warnings
import json
import shutil
import os
import inspect
import numpy as np
from pathlib import Path, PurePath


def get_export_function(output_name):
    r""" Get the appropriate export function for the chosen format.

        The function automatically detects the output format from the filename 
        extension,

         * '.pb' or '.PB': protobuf
         * '.kt' or '.KT': ketos
         * '.ktpb' or '.KTPB': ketos-protobuf (compatible with PAMGuard)

        For any other extension, the ketos export function is used.

        Args:
            output_name: str
                The name of the exported model.
        
        Returns:
            : func
                The export function
    """
    if output_name[-3:].lower() == '.pb':
        return export_to_protobuf
    elif output_name[-3:].lower() == '.kt':
        return export_to_ketos
    elif output_name[-5:].lower() == '.ktpb':
        return export_to_ketos_protobuf
    else:
        return export_to_ketos


def export_to_ketos_protobuf(model, output_name, audio_repr, input_shape=None,
                            tmp_folder="tmp_export_folder", overwrite=True, 
                            duration=None, backward_compat=True, **kwargs):
    r""" Export a ketos model to Ketos-Protobuf format. 

        Saving your ketos model in Ketos-Protobuf format makes it easier to share it with 
        collaborators and use it with other software applications.

        In particular, the output file generated by this function can be loaded directly into 
        PAMGuard, an open-source and widely adopted application for passive acoustic monitoring (PAM).

        If the output directory does not already exist, it is automatically created.
       
        The function generates a zipped archive containing,
        
         * the tensorflow model in protobuf format (model/model.pb)
         * the audio representation (audio_repr.json)
         * the ketos model recipe (recipe.json)

        The user is free to specify the extension of the output file, but we recommend using \*.ktpb
        as this will allow the file to be recognized and loaded into PAMGuard.

        A warning will be printed if the method is unable to infer values for the parameters 
        `duration` and `step`. The model will be saved, but it will not be possible to load 
        it into PAMGuard. 

        Args:
            model: 
                The ketos model to be exported. Usually created by one of the Interface classes found 
                in ketos.neural_networks (e.g.: ResNetInterface)
            output_name: str
                The name of the exported model. Must have the extension \*.ktpb to ensure that it can 
                be loaded into PAMGuard.
            input_shape: list or tuple.
                The input shape expected by the model. It can be represented by a tuple or list of four 
                elements: (number of intances, width, height, number of channels). The number of instances 
                and number of channels are commonly 1, and the width and height are usually the number of 
                time and frequency bins in a spectrogram, respectively. This, however, can vary with 
                the model in question.
            audio_repr: dict or str
                Audio representation. For example,

                >>> audio_repr = {"type": "MagSpectrogram",
                ...               "rate": "1000 Hz", 
                ...               "window": "0.256 s",
                ...               "step": "0.032 s",
                ...               "freq_min": "0 Hz",
                ...               "freq_max": "500 Hz",
                ...               "window_func": "hamming",
                ...               "transforms": [{"name":"normalize"}]
                ...              }                

                It is also possible to specify the path to a json file containing the audio representation.
            tmp_folder: str
                The name of a temporary folder created during the model conversion. It will be deleted 
                upon sucessful execution. If the folder already exists, a 'FileExistsError will be thrown, 
                unless 'overwrite' is set to True.
            overwrite: bool    
                If True and the folder specified in 'tmp_folder' exists, the folder will be overwritten.
            duration: float
                Duration in seconds of the input sample. If not specified, the duration is extracted from 
                the audio representation, or, if not available there, it is computed as `step * input_shape[0]`,
                provided that `input_shape` has been specified.
            backward_compat: bool
                Ensure backward compatibility with ketos versions 4.2.1 and older 

        Raises:
            AssertionError if the input shape cannot be inferred.
    """

    if input_shape != None and len(input_shape) == 4 and input_shape[0] == 1 and backward_compat:
        print("Warning: ketos is using the following interpretation of the input_shape:" \
                + f"{input_shape} = (number of intances, width, height, number of channels). " \
                + "If this is not the correct interpretation, you should call the " \
                + "export_to_ketos_protobuf function with backward_compat=False.")
        input_shape = input_shape[1:]

    # if the path to a json file containing the audio representation(s) 
    # has been specified, load its contents
    if isinstance(audio_repr, (str, PurePath)):
        audio_repr = pa.load_audio_representation(audio_repr, return_unparsed=True)
        
    if os.path.exists(tmp_folder):
        if not overwrite:
            raise FileExistsError("{} already exists. If you want to overwrite it set the \
                'overwrite' argument to True.".format(tmp_folder))
        else:
            shutil.rmtree(tmp_folder)

    # ensure that the audio representation is specified as a nested directory
    if 'type' in audio_repr.keys():
        audio_repr = {'spectrogram': audio_repr}

    # check that one of the representations is named 'spectrogram'
    # if this is not the case, issue a warning
    if 'spectrogram' not in audio_repr.keys():        
        warnings.warn("audio representation named `spectrogram` not found. You will not be " \
            "able to load the saved model into PAMGuard.", category=UserWarning)

    # if the output directory does not already exist, create it
    ensure_dir(output_name)

    # loop over audio representations
    for name, params in audio_repr.items():

        # parse audio representation, if it is not already parsed
        if pa.is_encoded(params):
            params = pa.parse_audio_representation(params)

        # if the user did not specify the sample duration, attempt to infer it 
        if duration == None:
            duration = _infer_duration(params, input_shape)
            if duration != None:
                print(f'Inferred sample duration of {duration}s')

        if duration == None:
            warnings.warn("ketos was unable to infer the sample duration. You will not be " \
                "able to load the saved model into PAMGuard. If you intend to load your model " \
                "into PAMGuard, you must specify the sample duration.", category=UserWarning)
        else:
            params['duration'] = duration    

        # if the user did not specify an input shape, attempt to infer it 
        if input_shape == None:
            input_shape = _infer_shape(params)
            if input_shape != None:
                print(f'Inferred input shape of {input_shape}')

        assert input_shape != None, 'ketos was unable to infer the input shape. Use the input_shape ' \
                                     + 'argument to specify the input shape expected by the model.'

        if 'step' not in params.keys():
            warnings.warn("ketos was unable to infer the step size. You will not be " \
                "able to load the saved model into PAMGuard. If you intend to load your model " \
                "into PAMGuard, you must specify the step size.", category=UserWarning)

        # if model has not been built, build it
        X = model._transform_input(np.ones(shape=input_shape))
        if not model.model.built:
            model.run_on_instance(X[0])

        params['dtype'] = model.model.dtype
        params['input_ndims'] = model.model.layers[0].input_spec.min_ndim
        params['input_shape'] = X.shape

        # encode audio representation parameters before saving to file
        params = pa.encode_audio_representation(params)

    # create tmp folder
    os.makedirs(tmp_folder)

    # save recipe to tmp folder
    recipe_path = os.path.join(tmp_folder, 'recipe.json')
    model.save_recipe_file(recipe_path)

    # save model to tmp folder 
    model_path = os.path.join(tmp_folder, 'model')
    save_pb(obj=model.model, export_dir=model_path)

    # save audio representation to tmp folder
    audio_repr_path = os.path.join(tmp_folder, "audio_repr.json")
    with open(audio_repr_path, 'w') as json_repr:
        json.dump(audio_repr, json_repr)

    # now, archive the contents of the tmp folder
    with ZipFile(output_name, 'w') as zip:
        zip.write(model_path, "model")
        for root, dirs, files in os.walk(model_path):
            renamed_root = root.replace(model_path, "model")
            for d in dirs:
                zip.write(os.path.join(root,d), os.path.join(renamed_root,d))
            for f in files:
                zip.write(os.path.join(root,f),os.path.join(renamed_root,f))            
                    
        zip.write(recipe_path, "recipe.json")
        zip.write(audio_repr_path, "audio_repr.json")
        
    # tidy up
    shutil.rmtree(tmp_folder)


def _infer_duration(params, input_shape):
    """ Helper function for :func:`export_to_ketos_protobuf`.

        The function first searches the `params` dictionary for an item 
        with key `duration`. If found, the value of this item is returned. 
        If not found, the function searches for the key `step` and, if found, 
        returns the product `params['step'] * input_shape[0]`. 
        If the function is unable to infer the duration through one of 
        these two approaches, it returns a None value.

        Args:
            params: dict
                Parsed audio representation.
            input_shape: tuple or list
                Shape of an input sample. 

        Returns:
            : float
                Duration in seconds.
    """
    if 'duration' in params.keys(): 
        return params['duration']

    elif input_shape != None and 'step' in params.keys():
        return params['step'] * input_shape[0]

    else:
        return None


def _infer_shape(params):
    """ Helper function for :func:`export_to_ketos_protobuf`.

        Args:
            params: dict
                Parsed audio representation.

        Returns:
            : tuple
                Inferred shape. If the shape could not be inferred,
                a None value is returned.
    """
    return params['type'].infer_shape(**params)


def export_to_protobuf(model, output_name, **kwargs):
    r""" Export a ketos model to Protobuf format (\*.pb).

        See also the related fuction :func:`export_to_ketos_protobuf`.

        Note that the model must be built before it can be saved.

        Args:
            model: 
                The ketos model to be exported. Usually created by one of the Interface 
                classes found in ketos.neural_networks (e.g.: ResNetInterface)
            output_name: str
                Folder where the exported model will be saved
    """
    assert model.model.built, "The model must be built. Call model.run_on_instance() on a sample input"

    # if the output directory does not already exist, create it
    ensure_dir(output_name)

    save_pb(obj=model.model, export_dir=output_name)


def export_to_ketos(model, output_name, checkpoint_name="cp-0000", audio_repr=None, custom_module=None, 
                 tmp_folder="tmp_export_folder", overwrite=True, metadata=None, extra=None, **kwargs):
    r""" Export a ketos model to ketos format (\*.kt).

        If the output directory does not already exist, it is automatically created.

        The function generates a zipped archive containing,

         * the model weights (\*.ckpt)
         * the audio representation (audio_repr.json)
         * the ketos model recipe (recipe.json)

        Args:
            model: 
                The ketos model to be exported. Usually created by one of the Interface classes found 
                in ketos.neural_networks (e.g.: ResNetInterface)
            output_name: str
                The name of the exported model. Should have the extension \*.kt
            checkpoint_name: str
                The name of the checkpoint file that the model weights will be saved to. Optional.
            audio_repr: dict or str
                Optional audio representation dictionary. If passed, it will be added to the \*.kt file.
                For example,
                    
                >>> audio_repr = {"spectrogram": {
                ...                   "type": "MagSpectrogram",
                ...                   "rate": "1000 Hz", 
                ...                   "window": "0.256 s",
                ...                   "step": "0.032 s",
                ...                   "freq_min": "0 Hz",
                ...                   "freq_max": "500 Hz",
                ...                   "window_func": "hamming",
                ...                   "transforms": [{"name":"normalize"}]}
                ...              }    

                It is also possible to specify the path to a json file containing the audio representation.
            custom_module: str
                Path to a custom module. The custom module can contain anynumber of files. Common files are user defined nn architecture, 
                audio representation, input transform function and output transform function. 
                All files inside the module will be saved to the ketos output file and can be later accessed to build the model for
                re-training or inference.
            tmp_folder: str
                The name of a temporary folder created during the model conversion. It will be deleted 
                upon sucessful execution. If the folder already exists, a 'FileExistsError will be thrown, 
                unless 'overwrite' is set to True.
            overwrite: bool    
                If True and the folder specified in 'tmp_folder' exists, the folder will be overwritten.
            metadata: dict
                Optional metadata dictionary. If passed, it will be added to the \*.kt output archive file. 
            extra: str, list(str)
                Full path to one or several additional files to be saved to the output \*.kt archive file.
    """

    if not isinstance(extra, list) and extra is not None:
        extra = [extra]

    tmp_folder = Path(tmp_folder)
    # create tmp folder
    if tmp_folder.exists():
        if not overwrite:
            raise FileExistsError("{} already exists. If you want to overwrite it set the \
                'overwrite' argument to True.".format(tmp_folder))
        else:
            shutil.rmtree(tmp_folder)

    recipe = model._extract_recipe_dict()

    # save audio representation to tmp folder
    if audio_repr is not None:
        if isinstance(audio_repr, str) and Path(audio_repr).suffix == ".json":
            audio_repr = pa.load_audio_representation(audio_repr, return_unparsed=False)

        if isinstance(audio_repr, dict) and pa.is_encoded(audio_repr):
            audio_repr = pa.parse_audio_representation(audio_repr)

        audio_repr = pa.encode_audio_representation(audio_repr)

    if custom_module is not None:
        custom_module = Path(custom_module).resolve()

    # save checkpoints to tmp folder 
    # Unfortunetly, the current implementation of save_eights and the ZipFile library does not allow to save the model weights directly to the zipfile
    weights_path = (tmp_folder / "checkpoints" / checkpoint_name)
    weights_path.parent.mkdir(parents=True, exist_ok=True)
    model.model.save_weights(weights_path) 

    # save to zip archive
    with ZipFile(output_name, 'w') as archive:
        
        # save recipe 
        recipe = json.dumps(recipe) # converts the dictionary as a json_string
        archive.writestr('recipe.json', recipe)

        # save audio representation
        if audio_repr is not None:
            audio_repr = json.dumps(audio_repr) # properly converts to json string
            archive.writestr('audio_repr.json', audio_repr)

        # save metadata
        if metadata is not None:
            metadata = json.dumps(metadata)
            archive.writestr("metadata.json", metadata)

        # save checkpoints
        checkpoints = weights_path.parent.glob('*')
        for checkpoint in checkpoints:
            archive.write(checkpoint, Path("checkpoints", checkpoint.name))

        # Save custom module to the archive
        if custom_module is not None:
            files = custom_module.rglob('*')
            for file in files:
                archive.write(file, "custom/" + file.name)
    
        # save any extra contents  
        if extra is not None:   
            for extra_file in extra:
                archive.write(extra_file, Path(extra_file).name)

    # tidy up
    shutil.rmtree(tmp_folder)



from ketos.neural_networks.cnn import CNNInterface, CNN1DInterface
from ketos.neural_networks.resnet import ResNetInterface, ResNet1DInterface
from ketos.neural_networks.densenet import DenseNetInterface
from ketos.neural_networks.inception import InceptionInterface
from zipfile import ZipFile
from pathlib import Path
import warnings
import importlib
import sys
import zipimport
import json


interface_names_in_recipes = {'CNNInterface':CNNInterface,
                              'CNN1DInterface':CNN1DInterface,
                              'ResNetInterface':ResNetInterface,
                              'ResNet1DInterface':ResNet1DInterface,
                              'DenseNetInterface':DenseNetInterface,
                              'InceptionInterface':InceptionInterface}

def load_model_file(model_file, new_model_folder='kt-tmp', overwrite=True, load_audio_repr=False, replace_top=False, diff_n_classes=None):
    """ Load a model from a ketos (.kt) model file.
        
        Args:
            model_file:str
                Path to the ketos(.kt) file
            new_model_folder:str
                Path to folder where files associated with the model will be stored.
                By default the files will be saved to a folder named 'kt-tmp' created 
                in the current working directory.
            overwrite: bool
                If True, the 'new_model_folder' will be overwritten.
            replace_top: bool
                If True, the classification top of the model will be replaced by a new, untrained one.
                What is actually replaced (i.e.: what exactly is the "top") is defined by the architecture.
                It is usually a block of dense layers with the appropriate activations. Default is False.
            diff_n_classes: int
                Only relevant when 'replace_top' is True.
                If the new model should have a different number of classes it can be specified by this parameter.
                If left to none, the new model will have the same number of classes as the original.
            load_audio_repr: bool
                If True, look for an audio representation included with the model. 
                
        Raises:
            ValueError: If the model recipe does not contain a valid value for the 'interface' field.

        Returns:
            model_instance: The loaded model
            audio_repr: If load_audio_repr is True, also return a dictionary with the loaded audio representation.

    """
    # First we need to get the recipe file to know which Interface we are working with.
    path = Path(model_file).resolve()
    with ZipFile(path, 'r') as archive: # we dont need to extract from the zipfile, we can just load directly into memory
        recipe_file = archive.open('recipe.json')
        nn_recipe = json.load(recipe_file)
    
    # The script prioritizes custom architectrues over the default ones from ketos. So if the user defined a custom architecture with the same name as a default one, it will prioritize the custom
    # if there is no custom_architecture it will attempt to load a default one
    try:
        # See docs https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        module_name = "nn_architecture" #The file name will always be custom_architecture because this is how ketos saves it
        module_path = zipimport.zipimporter(path / "custom") # the folder name within the zip archive will always be custom
        spec = module_path.find_spec(module_name)

        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module # adding the module to the list of modules in sys
        spec.loader.exec_module(module) # actually load the module, this is equivalent to importing
        # Now we load the class
        interface = getattr(module, nn_recipe['interface'])
        # Warn the user about using a custom input transform function
        warnings.warn("Ketos detected a custom nn architecture and will be using that.", UserWarning)
    except AttributeError:
        # fallback to the dictionary
        try:
            # load the class from the default interfaces in ketos
            interface = interface_names_in_recipes[nn_recipe['interface']]
        except KeyError as e:
            e.__suppress_context__ = True # we are supressing the first attribute error as it is not useful
            raise ValueError("The model recipe does not indicate a valid interface")
    

    # load the model with the appropriate interface
    loaded = interface.load(model_file=model_file, new_model_folder=new_model_folder, overwrite=overwrite, load_audio_repr=load_audio_repr,  replace_top=replace_top, diff_n_classes=diff_n_classes)
    return loaded

def load_input_transform_function(model_file, function_name="transform"):
    """ Load a input transform function from a ketos (.kt) model file.
        
        Args:
            model_file:str
                Path to the ketos(.kt) file

        Returns:
            function: The input transform function

    """
    # loads a custom input transform function from the model file.
    # Gives an Attribute error if there is none

    path = Path(model_file).resolve()

    module_name = "input_transform_function" #The file name will always be custom_architecture because this is how ketos saves it
    module_path = zipimport.zipimporter(path / "custom") # the folder name within the zip archive will always be custom
    spec = module_path.find_spec(module_name)

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module # adding the module to the list of modules in sys
    spec.loader.exec_module(module) # actually load the module, this is equivalent to importing
    # Now we load the function
    function = getattr(module, function_name)

    return function

def load_export_transform_function(model_file, function_name="transform"):
    """ Load an export transform function from a ketos (.kt) model file.
        
        Args:
            model_file:str
                Path to the ketos(.kt) file

        Returns:
            function: The export transform function

    """
    # loads a custom input transform function frrom the model file if there is one
    # Otherwise giver error

    path = Path(model_file).resolve()

    module_name = "output_transform_function" #The file name will always be custom_architecture because this is how ketos saves it
    module_path = zipimport.zipimporter(path / "custom") # the folder name within the zip archive will always be custom
    spec = module_path.find_spec(module_name)

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module # adding the module to the list of modules in sys
    spec.loader.exec_module(module) # actually load the module, this is equivalent to importing
    # Now we load the function
    function = getattr(module, function_name)

    return function



import numpy as np
np.random.seed(1000)

import tensorflow as tf
tf.random.set_seed(2000)



import pytest
from ketos.neural_networks.dev_utils.nn_interface import RecipeCompat, NNInterface, NNArch
from ketos.neural_networks.dev_utils.losses import FScoreLoss
from ketos.data_handling.data_feeding import BatchGenerator
from unittest.mock import Mock, patch
import os
import shutil
import tables
import json
from zipfile import ZipFile


current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')


@pytest.fixture
def batch_generator():
    data = np.vstack([np.zeros((10,512)), np.ones((10,512))])
    labels = np.concatenate([np.array([[1,0] for i in range(10)]), np.array([[0,1] for i in range(10)])])
    generator = BatchGenerator(batch_size=5, x=data, y=labels, shuffle=True)

    return generator


@pytest.fixture
def recipe_dict():
    recipe = {'optimizer': {'recipe_name':'Adam', 'parameters': {'learning_rate':0.005}},
               'loss_function': {'recipe_name':'FScoreLoss', 'parameters':{}},  
               'metrics': [{'recipe_name':'CategoricalAccuracy', 'parameters':{}}]
        
    }
    return recipe

@pytest.fixture
def MLPInterface_subclass():
    """ A simple MLP inheriting from NNInterface
    """
    
    class MLP(NNArch):
        def __init__(self, n_neurons, activation):
            super(MLP, self).__init__()

            self.dense = tf.keras.layers.Dense(n_neurons, activation=activation)
            self.final_node = tf.keras.layers.Dense(2, 'softmax')

        def call(self, inputs):
            output = self.call_frontend(inputs)
            output = self.dense(output)
            output = self.final_node(output)
            return output

            
    class MLPInterface(NNInterface):

        @classmethod
        def _transform_input(cls, input):
            if input.ndim == 1:
                transformed_input = input.reshape(1,input.shape[0])
            elif input.ndim == 2:
                transformed_input = input.reshape(input.shape[0],input.shape[1])
            else:
                raise ValueError("Expected input to have 1 or 2 dimensions, got {}({}) instead".format(input.ndims, input.shape))

            return transformed_input

        @classmethod
        def _transform_output(cls,output):
            max_class = np.argmax(output, axis=-1)
            if output.shape[0] == 1:
                max_class_conf = output[0][max_class]
                transformed_output = (max_class[0], max_class_conf[0])
            elif output.shape[0] > 1:
                max_class_conf = np.array([output[i][c] for i, c in enumerate(max_class)])

            transformed_output = (max_class, max_class_conf)
            
            return transformed_output



        @classmethod
        def build_from_recipe(cls, recipe):
            n_neurons = recipe['n_neurons']
            activation = recipe['activation']
            optimizer = recipe['optimizer']
            loss_function = recipe['loss_function']
            metrics = recipe['metrics']

            instance = cls(n_neurons=n_neurons, activation=activation, optimizer=optimizer, loss_function=loss_function, metrics=metrics, secondary_metrics=secondary_metrics)

            return instance

        @classmethod
        def _read_recipe_file(cls, json_file, return_recipe_compat=True):
            
            with open(json_file, 'r') as json_recipe:
                recipe_dict = json.load(json_recipe)
            
            
            optimizer = cls._optimizer_from_recipe(recipe_dict['optimizer'])
            loss_function = cls._loss_function_from_recipe(recipe_dict['loss_function'])
            metrics = cls._metrics_from_recipe(recipe_dict['metrics'])
            

            if return_recipe_compat == True:
                recipe_dict['optimizer'] = optimizer
                recipe_dict['loss_function'] = loss_function
                recipe_dict['metrics'] = metrics
            
            else:
                recipe_dict['optimizer'] = cls._optimizer_to_recipe(optimizer)
                recipe_dict['loss_function'] = cls._loss_function_to_recipe(loss_function)
                recipe_dict['metrics'] = cls._metrics_to_recipe(metrics)
            
            recipe_dict['n_neurons'] = recipe_dict['n_neurons']
            recipe_dict['activation'] = recipe_dict['activation']
            return recipe_dict

        def __init__(self, n_neurons, activation, optimizer, loss_function, metrics):
            super(MLPInterface, self).__init__(optimizer, loss_function, metrics)
            self.n_neurons = n_neurons
            self.activation = activation
            self.model = MLP(n_neurons=n_neurons, activation=activation)
            

        def _extract_recipe_dict(self):
        
            recipe = {}
            recipe['optimizer'] = self._optimizer_to_recipe(self.optimizer)
            recipe['loss_function'] = self._loss_function_to_recipe(self.loss_function)
            recipe['metrics'] = self._metrics_to_recipe(self.metrics)
            recipe['n_neurons'] = self.n_neurons
            recipe['activation'] = self.activation

            return recipe

    return MLPInterface


@pytest.fixture
def instance_of_MLPInterface(MLPInterface_subclass):
    np.random.seed(1000)
    tf.random.set_seed(2000)
    path_to_file = os.path.join(path_to_assets, "recipes/basic_recipe.json")
    recipe = NNInterface._read_recipe_file(path_to_file)

    h5 = tables.open_file(os.path.join(path_to_assets, "vectors_1_0.h5"), 'r')
    train_table = h5.get_node("/train")
    val_table = h5.get_node("/val")
    test_table = h5.get_node("/test")

    data = np.vstack([np.zeros((10,512)), np.ones((10,512))])
    labels = np.concatenate([np.array([[1,0] for i in range(10)]), np.array([[0,1] for i in range(10)])])
    train_generator = BatchGenerator(batch_size=5, x=data, y=labels, shuffle=True)
    val_generator = BatchGenerator(batch_size=5, x=data, y=labels, shuffle=True)
    test_generator = BatchGenerator(batch_size=5, x=data, y=labels, shuffle=True)
    

    instance = MLPInterface_subclass(activation='relu', n_neurons=64, optimizer=recipe['optimizer'],
                         loss_function=recipe['loss_function'], metrics=recipe['metrics']) 

    instance.train_generator = train_generator
    instance.val_generator = val_generator
    instance.test_generator = test_generator
    
    return instance


def test_RecipeCompat():
    opt = RecipeCompat("sgd", tf.keras.optimizers.SGD, learning_rate=0.008, momentum=0.1)
    assert opt.recipe_name == "sgd"
    assert opt.args ==  {'learning_rate': 0.008, 'momentum': 0.1}
    assert isinstance(opt.instance, tf.keras.optimizers.SGD) 
    assert opt.instance.learning_rate == 0.008
    assert opt.instance.momentum == 0.1

    metric = RecipeCompat("accuracy", tf.keras.metrics.Accuracy, name='acc')
    assert metric.recipe_name == "accuracy"
    assert metric.args == {'name': 'acc'}
    assert isinstance(metric.instance, tf.keras.metrics.Accuracy) 
    assert metric.instance.name == 'acc'

 

@pytest.mark.parametrize("class_label,n_classes,expected",[
    (0,2,np.array([1.0,0.0])),
    (1,2,np.array([0.0,1.0])),
    (1,3,np.array([0.0,1.0, 0.0])),
    (1,5,np.array([0.0,1.0, 0.0, 0.0, 0.0])),
    ])
def test_to1hot(class_label, n_classes, expected):
    encoded =  NNInterface._to1hot(class_label, n_classes) 
    assert(encoded == expected).all()
    with pytest.raises(IndexError):
        assert NNInterface._to1hot(class_label=1.2, n_classes=2) 
    with pytest.raises(TypeError):
        assert NNInterface._to1hot(class_label=1, n_classes=2.0)



def test_transform_batch():
    inputs = np.random.rand(10,5,5)
    labels = np.array([1,0,0,0,1,0,0,1,1,1])

    X,Y = NNInterface.transform_batch(inputs, labels)

    
    assert X.shape == (10,5,5,1)
    assert np.array_equal(X[:,:,:,0], inputs)
    assert np.array_equal(Y, np.array([[0., 1.],
                            [1., 0.],
                            [1., 0.],
                            [1., 0.],
                            [0., 1.],
                            [1., 0.],
                            [1., 0.],
                            [0., 1.],
                            [0., 1.],
                            [0., 1.]]))


def test_transform_input():
    input1 = np.random.rand(5,5)        
    input2 = np.random.rand(1,5,5)

    output1 = NNInterface._transform_input(input1)
    output2 = NNInterface._transform_input(input2)

    assert output1.shape == (1,5,5,1)
    assert output2.shape == (1,5,5,1)

    assert np.array_equal(output1[0,:,:,0], input1)
    assert np.array_equal(output2[0,:,:,0], input2[0,:,:])


def test_transform_output():
    output1 = np.array([[0.05,0.05,0.7, 0.1, 0.1]])
    output2 = np.array([[0.05,0.05,0.7, 0.1, 0.1],[0.05,0.15,0.15,0.65,0.1]])

    class1, score1 = NNInterface._transform_output(output1)
    classes2, scores2 = NNInterface._transform_output(output2)

    assert np.array_equal(class1, np.array([2])) 
    assert np.array_equal(score1, np.array([0.7]))
    
    assert np.array_equal(classes2, np.array([2, 3])) 
    assert np.array_equal(scores2, np.array([0.7, 0.65]))
    


def test_optimizer_from_recipe(recipe_dict):
    built_opt = NNInterface._optimizer_from_recipe(recipe_dict['optimizer'])
    assert isinstance(built_opt, RecipeCompat)
    assert built_opt.recipe_name == 'Adam'
    assert built_opt.args == {'learning_rate':0.005}
    assert isinstance(built_opt.instance, tf.keras.optimizers.Adam)
    
    
def test_loss_function_from_recipe(recipe_dict):
    built_loss = NNInterface._loss_function_from_recipe(recipe_dict['loss_function'])
    assert isinstance(built_loss, RecipeCompat)
    assert built_loss.recipe_name == 'FScoreLoss'
    assert built_loss.args == {}
    assert isinstance(built_loss.instance, FScoreLoss)
    
def test_metrics_from_recipe(recipe_dict):
    built_metrics = NNInterface._metrics_from_recipe(recipe_dict['metrics'])
    assert isinstance(built_metrics[0], RecipeCompat)
    assert built_metrics[0].recipe_name == 'CategoricalAccuracy'
    assert built_metrics[0].args == {}
    assert isinstance(built_metrics[0].instance, tf.keras.metrics.CategoricalAccuracy)
    
    
def test_optimizer_to_recipe():
    recipe_compat_opt = RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005)
    optimizer_dict = NNInterface._optimizer_to_recipe(recipe_compat_opt)

    assert optimizer_dict['recipe_name'] == 'Adam'
    assert optimizer_dict['parameters'] == {'learning_rate':0.005}


def test_loss_function_to_recipe():
    recipe_compat_loss = RecipeCompat('FScoreLoss', FScoreLoss, beta=0.5)
    loss_dict = NNInterface._loss_function_to_recipe(recipe_compat_loss)

    assert loss_dict['recipe_name'] == 'FScoreLoss'
    assert loss_dict['parameters'] == {'beta':0.5}

    
def test_metrics_to_recipe():
    recipe_compat_metric = [RecipeCompat('CategoricalAccuracy', tf.keras.metrics.CategoricalAccuracy)]
    metrics_dicts = NNInterface._metrics_to_recipe(recipe_compat_metric)

    assert isinstance(metrics_dicts, list)
    assert metrics_dicts[0]['recipe_name'] == 'CategoricalAccuracy'
    assert metrics_dicts[0]['parameters'] == {}


def test_read_recipe_file():
    path_to_file = os.path.join(path_to_assets, "recipes/basic_recipe.json")
    recipe = NNInterface._read_recipe_file(path_to_file)

    assert isinstance(recipe, dict)
    assert isinstance(recipe['optimizer'], RecipeCompat)
    assert isinstance(recipe['loss_function'], RecipeCompat)
    assert isinstance(recipe['metrics'], list)
    assert isinstance(recipe['metrics'][0], RecipeCompat)

    opt = recipe['optimizer']
    assert opt.recipe_name == 'Adam'
    assert isinstance(opt.instance, tf.keras.optimizers.Adam)
    assert opt.args == {"learning_rate": 0.001, "beta_1": 0.9, "beta_2": 0.999, "decay": 0.01}

    loss = recipe['loss_function']
    assert loss.recipe_name == 'FScoreLoss'
    assert isinstance(loss.instance, FScoreLoss)
    assert loss.args == {}

    metric = recipe['metrics'][0]
    assert metric.recipe_name == 'CategoricalAccuracy'
    assert isinstance(metric.instance, tf.keras.metrics.CategoricalAccuracy)
    assert metric.args == {}


def test_write_recipe_file(recipe_dict):
    destination = os.path.join(path_to_tmp, "test_write_recipe_file.json")
    NNInterface._write_recipe_file(destination, recipe_dict)

    read_recipe =  NNInterface._read_recipe_file(destination, return_recipe_compat=False)
    #If return_recipe_compat is False, the result should be a recipe dictionary just like the recipe_dict used to write the file
    assert read_recipe == recipe_dict


    read_recipe =  NNInterface._read_recipe_file(destination, return_recipe_compat=True)
    #If return_recipe_compat is True, the result will be a dictionary with RecipeCompat objects for the optimizer, loss_functions and metrics (actually a lis of RecipCompat objects) 

    assert read_recipe['optimizer'].recipe_name == NNInterface._optimizer_from_recipe(recipe_dict['optimizer']).recipe_name
    assert read_recipe['optimizer'].instance.__class__ == NNInterface._optimizer_from_recipe(recipe_dict['optimizer']).instance.__class__
    assert read_recipe['optimizer'].args == NNInterface._optimizer_from_recipe(recipe_dict['optimizer']).args

    assert read_recipe['loss_function'].recipe_name == NNInterface._loss_function_from_recipe(recipe_dict['loss_function']).recipe_name
    assert read_recipe['loss_function'].instance.__class__ == NNInterface._loss_function_from_recipe(recipe_dict['loss_function']).instance.__class__
    assert read_recipe['loss_function'].args == NNInterface._loss_function_from_recipe(recipe_dict['loss_function']).args
    
    assert read_recipe['metrics'][0].recipe_name == NNInterface._metrics_from_recipe(recipe_dict['metrics'])[0].recipe_name
    assert read_recipe['metrics'][0].instance.__class__ == NNInterface._metrics_from_recipe(recipe_dict['metrics'])[0].instance.__class__
    assert read_recipe['metrics'][0].args == NNInterface._metrics_from_recipe(recipe_dict['metrics'])[0].args
    

    
def test_instantiate_nn(MLPInterface_subclass):
    path_to_file = os.path.join(path_to_assets, "recipes/basic_recipe.json")
    recipe = NNInterface._read_recipe_file(path_to_file)
    
    MLPInterface_subclass(activation='relu', n_neurons=64, optimizer=recipe['optimizer'],
                         loss_function=recipe['loss_function'], metrics=recipe['metrics'])    

def test_save_recipe_file(instance_of_MLPInterface):
    path_to_file = os.path.join(path_to_assets, "recipes/basic_recipe.json")
    recipe = NNInterface._read_recipe_file(path_to_file)

    instance = instance_of_MLPInterface

    path_to_saved_recipe = os.path.join(path_to_tmp, "test_save_recipe.json")
    instance.save_recipe_file(path_to_saved_recipe)

    read_recipe = instance._read_recipe_file(path_to_saved_recipe)


    assert read_recipe['n_neurons'] == 64
    assert read_recipe['activation'] == 'relu'
    assert read_recipe['optimizer'].recipe_name ==recipe['optimizer'].recipe_name
    assert read_recipe['optimizer'].instance.__class__ == recipe['optimizer'].instance.__class__
    assert read_recipe['optimizer'].args == recipe['optimizer'].args

    assert read_recipe['loss_function'].recipe_name == recipe['loss_function'].recipe_name
    assert read_recipe['loss_function'].instance.__class__ == recipe['loss_function'].instance.__class__
    assert read_recipe['loss_function'].args == recipe['loss_function'].args
    
    assert read_recipe['metrics'][0].recipe_name == recipe['metrics'][0].recipe_name
    assert read_recipe['metrics'][0].instance.__class__ == recipe['metrics'][0].instance.__class__
    assert read_recipe['metrics'][0].args == recipe['metrics'][0].args


def test_extract_recipe_dict(instance_of_MLPInterface):
    recipe = instance_of_MLPInterface._extract_recipe_dict()
    
    assert recipe['n_neurons'] == 64
    assert recipe['activation'] == 'relu'

    assert recipe['optimizer'] == {'recipe_name':'Adam', 'parameters': {'beta_1': 0.9, 'beta_2': 0.999, 'decay': 0.01, 'learning_rate': 0.001}}
    assert recipe['loss_function'] == {'recipe_name':'FScoreLoss', 'parameters':{}}
    assert recipe['metrics'] == [{'recipe_name':'CategoricalAccuracy', 'parameters':{}}]


def test_train_loop_early_stop_metric_decrease(instance_of_MLPInterface):
   
    train_loss_values = iter([0.9,0.9,0.9,0.9,
                              0.8,0.8,0.8,0.8,
                              0.7,0.7,0.7,0.7,
                              0.7,0.7,0.7,0.7,
                              0.7,0.7,0.7,0.7,
                              0.7,0.7,0.7,0.7,
                              0.7,0.7,0.7,0.7,
                              0.7,0.7,0.7,0.7,
                              0.6,0.6,0.6,0.6,
                              0.5,0.5,0.5,0.5,
                              0.4,0.4,0.4,0.4,
                              0.3,0.3,0.3,0.3,
                              0.2,0.2,0.2,0.2,
                              0.1,0.1,0.1,0.1,])

    def mock_train_step(self, inputs, labels):
       self._train_loss = Mock()
       loss = next(train_loss_values)
       self._train_loss.result.return_value = loss#next(train_loss_values)

    patcher = patch('ketos.neural_networks.dev_utils.nn_interface.NNInterface._train_step', new=mock_train_step)
    patcher.start()
    
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_checkpoints")
    

    instance_of_MLPInterface._early_stopping_monitor = {"metric": 'train_loss',
                                        "decreasing": True,
                                        "period":3,
                                        "min_epochs": 3,
                                        "max_epochs": None,
                                        "delta" : 0.1,
                                        "baseline":0.5}
    
    instance_of_MLPInterface.train_loop(n_epochs=8, early_stopping=True)
    assert instance_of_MLPInterface.last_epoch_with_improvement == 3
 
    patcher.stop()
    
def test_train_loop_early_stop_metric_decrease_baseline(instance_of_MLPInterface):
   
    train_loss_values = iter([0.9,0.9,0.9,0.9,
                              0.8,0.8,0.8,0.8,
                              0.7,0.7,0.7,0.7,
                              0.7,0.7,0.7,0.7,
                              0.5,0.5,0.5,0.5,
                              0.4,0.4,0.4,0.4,
                              0.3,0.3,0.3,0.3,
                              0.2,0.2,0.2,0.2,
                              0.1,0.1,0.1,0.1,])

    def mock_train_step(self, inputs, labels):
       self._train_loss = Mock()
       loss = next(train_loss_values)
       self._train_loss.result.return_value = loss#next(train_loss_values)

    patcher = patch('ketos.neural_networks.dev_utils.nn_interface.NNInterface._train_step', new=mock_train_step)
    patcher.start()
    
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_checkpoints")

    instance_of_MLPInterface._early_stopping_monitor = {"metric": 'train_loss',
                                        "decreasing": True,
                                        "period":3,
                                        "min_epochs": 3,
                                        "max_epochs": None,
                                        "delta" : 0.1,
                                        "baseline":0.5}
    
    instance_of_MLPInterface.train_loop(n_epochs=8, early_stopping=True)
    assert instance_of_MLPInterface.last_epoch_with_improvement == 4
 
    patcher.stop()
    
def test_train_loop_early_stop_metric_increase(instance_of_MLPInterface):
   
    train_loss_values = iter([0.1,0.1,0.1,0.1,
                             0.2,0.2,0.2,0.2,
                             0.3,0.3,0.3,0.3,
                             0.3,0.3,0.3,0.3,
                             0.3,0.3,0.3,0.3,
                             0.3,0.3,0.3,0.3,
                             0.3,0.3,0.3,0.3,
                             0.3,0.3,0.3,0.3,
                             0.3,0.3,0.3,0.3,
                             0.3,0.3,0.3,0.3,
                             0.4,0.4,0.4,0.4,
                             0.4,0.4,0.4,0.4,
                             0.5,0.5,0.5,0.5,
                             0.6,0.6,0.6,0.6,
                                                        ])

    def mock_train_step(self, inputs, labels):
       self._train_loss = Mock()
       loss = next(train_loss_values)
       self._train_loss.result.return_value = loss#next(train_loss_values)

    patcher = patch('ketos.neural_networks.dev_utils.nn_interface.NNInterface._train_step', new=mock_train_step)
    patcher.start()
    
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_checkpoints")

    instance_of_MLPInterface._early_stopping_monitor = {"metric": 'train_loss',
                                        "decreasing": True,
                                        "period":3,
                                        "min_epochs": 3,
                                        "max_epochs": None,
                                        "delta" : 0.1,
                                        "baseline":1.0}
    
    instance_of_MLPInterface.train_loop(n_epochs=8, early_stopping=True)
    assert instance_of_MLPInterface.last_epoch_with_improvement == 3
 
    patcher.stop()
    
def test_train_loop(instance_of_MLPInterface):
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5)

def test_save_model_latest(instance_of_MLPInterface):
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_save_model_latest_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5, checkpoint_freq=1)
    instance_of_MLPInterface.save(os.path.join(path_to_tmp, "test_save_model_latest.kt"))
    extraction_path = os.path.join(path_to_tmp, "test_save_model_latest_tmp_folder")
    with ZipFile(os.path.join(path_to_tmp, "test_save_model_latest.kt"), 'r') as zip:
            zip.extractall(path=extraction_path)
    
    assert os.path.exists(os.path.join(extraction_path,"checkpoints", "cp-0005.ckpt.data-00000-of-00001" ))
    assert os.path.exists(os.path.join(extraction_path, "checkpoints", "cp-0005.ckpt.index"))
    assert os.path.exists(os.path.join(extraction_path, "checkpoints", "checkpoint"))

def test_save_model_latest_protobuf(instance_of_MLPInterface):
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_save_model_latest_checkpoints_protobuf")
    instance_of_MLPInterface.train_loop(n_epochs=5, checkpoint_freq=1)
    output_path = os.path.join(path_to_tmp, "test_save_model_latest.pb")
    instance_of_MLPInterface.save(output_path)
    assert os.path.exists(output_path)

def test_save_model_checkpoint_name(instance_of_MLPInterface):
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_save_model_latest_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5, checkpoint_freq=1)
    instance_of_MLPInterface.save(os.path.join(path_to_tmp, "test_save_model_latest.kt"), checkpoint_name="cp-0003.ckpt")
    extraction_path = os.path.join(path_to_tmp, "test_save_model_latest_tmp_folder")
    with ZipFile(os.path.join(path_to_tmp, "test_save_model_latest.kt"), 'r') as zip:
            zip.extractall(path=extraction_path)
    
    assert os.path.exists(os.path.join(extraction_path,"checkpoints", "cp-0003.ckpt.data-00000-of-00001" ))
    assert os.path.exists(os.path.join(extraction_path, "checkpoints", "cp-0003.ckpt.index"))
    assert os.path.exists(os.path.join(extraction_path, "checkpoints", "checkpoint"))

def test_save_model_with_audio_repr(instance_of_MLPInterface, spectr_settings):
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_save_model_latest_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5, checkpoint_freq=1)
    instance_of_MLPInterface.save(os.path.join(path_to_tmp, "test_save_model_latest.kt"), audio_repr=json.loads(spectr_settings))
    extraction_path = os.path.join(path_to_tmp, "test_save_model_latest_tmp_folder")
    with ZipFile(os.path.join(path_to_tmp, "test_save_model_latest.kt"), 'r') as zip:
            zip.extractall(path=extraction_path)
    
    assert os.path.exists(os.path.join(extraction_path,"checkpoints", "cp-0005.ckpt.data-00000-of-00001" ))
    assert os.path.exists(os.path.join(extraction_path, "checkpoints", "cp-0005.ckpt.index"))
    assert os.path.exists(os.path.join(extraction_path, "checkpoints", "checkpoint"))
    assert os.path.exists(os.path.join(extraction_path, "audio_repr.json"))

def test_save_model_with_audio_repr_file(instance_of_MLPInterface):
    audio_repr_file = os.path.join(path_to_assets, "audio_repr.json")
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_save_model_latest_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5, checkpoint_freq=1)
    instance_of_MLPInterface.save(os.path.join(path_to_tmp, "test_save_model_latest.kt"), audio_repr_file=audio_repr_file)
    extraction_path = os.path.join(path_to_tmp, "test_save_model_latest_tmp_folder")
    with ZipFile(os.path.join(path_to_tmp, "test_save_model_latest.kt"), 'r') as zip:
            zip.extractall(path=extraction_path)
    
    assert os.path.exists(os.path.join(extraction_path,"checkpoints", "cp-0005.ckpt.data-00000-of-00001" ))
    assert os.path.exists(os.path.join(extraction_path, "checkpoints", "cp-0005.ckpt.index"))
    assert os.path.exists(os.path.join(extraction_path, "checkpoints", "checkpoint"))
    assert os.path.exists(os.path.join(extraction_path, "audio_repr.json"))

def test_train_loop_log_csv(MLPInterface_subclass,batch_generator):
    recipe = { 'n_neurons':64,
               'activation':'relu',
               'optimizer': RecipeCompat("Adam",tf.keras.optimizers.Adam,learning_rate=0.005),
               'loss_function': RecipeCompat("FScoreLoss",FScoreLoss),  
               'metrics': [RecipeCompat('CategoricalAccuracy', tf.keras.metrics.CategoricalAccuracy),
                            RecipeCompat('Precision', tf.keras.metrics.Precision),
                            RecipeCompat('Recall', tf.keras.metrics.Recall)],
            }

    h5 = tables.open_file(os.path.join(path_to_assets, "vectors_1_0.h5"), 'r')
    train_table = h5.get_node("/train")
    val_table = h5.get_node("/val")
    test_table = h5.get_node("/test")

    instance = MLPInterface_subclass(activation='relu', n_neurons=64, optimizer=recipe['optimizer'],
                         loss_function=recipe['loss_function'], metrics=recipe['metrics']) 

    instance.train_generator = batch_generator
    instance.val_generator = batch_generator
    instance.log_dir = os.path.join(path_to_tmp, "test_log_dir")
    instance.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_log_csv_checkpoints")

    instance.train_loop(15, log_csv=True)

    assert os.path.isfile(os.path.join(instance.log_dir, "log.csv"))
    os.remove(os.path.join(instance.log_dir, "log.csv"))

def test_train_loop_log_tensorboard(MLPInterface_subclass):
    recipe = { 'n_neurons':64,
               'activation':'relu',
               'optimizer': RecipeCompat("Adam",tf.keras.optimizers.Adam,learning_rate=0.005),
               'loss_function': RecipeCompat("FScoreLoss",FScoreLoss),  
               'metrics': [RecipeCompat('CategoricalAccuracy', tf.keras.metrics.CategoricalAccuracy),
                            RecipeCompat('Precision', tf.keras.metrics.Precision),
                            RecipeCompat('Recall', tf.keras.metrics.Recall)],
               
            }

    h5 = tables.open_file(os.path.join(path_to_assets, "vectors_1_0.h5"), 'r')
    train_table = h5.get_node("/train")
    val_table = h5.get_node("/val")
    test_table = h5.get_node("/test")

    data = np.vstack([np.zeros((10,512)), np.ones((10,512))])
    labels = np.concatenate([np.array([[1,0] for i in range(10)]), np.array([[0,1] for i in range(10)])])
    train_generator = BatchGenerator(batch_size=5, x=data, y=labels, shuffle=True)
    val_generator = BatchGenerator(batch_size=5, x=data, y=labels, shuffle=True)
    test_generator = BatchGenerator(batch_size=5, x=data, y=labels, shuffle=True)
    
    instance = MLPInterface_subclass(activation='relu', n_neurons=64, optimizer=recipe['optimizer'],
                         loss_function=recipe['loss_function'], metrics=recipe['metrics']) 

    instance.train_generator = train_generator
    instance.val_generator = val_generator
    instance.log_dir = os.path.join(path_to_tmp, "test_log_dir")
    instance.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_log_tensorboard_checkpoints")

    instance.train_loop(15, log_tensorboard=True)

    assert os.path.isdir(os.path.join(instance.log_dir, "tensorboard_metrics"))
    shutil.rmtree(os.path.join(instance.log_dir, "tensorboard_metrics"))


def test_run_on_batch(instance_of_MLPInterface):
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5)
    instance_of_MLPInterface.test_generator.batch_count = 0 #reset batch generator
    X,y = next(instance_of_MLPInterface.test_generator)

    pred, scores = instance_of_MLPInterface.run_on_batch(X, transform_input=True, return_raw_output=False)
    
    assert pred.shape == (5,)
    assert np.array_equal(pred, np.array([0,1,1,0,0]))

    assert scores.shape == (5,)
    np.testing.assert_array_almost_equal(scores, np.array([0.5086525,1.,1., 0.5086525, 0.5086525], dtype=scores.dtype), decimal=7)

def test_frontend_layer(instance_of_MLPInterface):
    frontend_layer = tf.keras.layers.Dense(64, activation='relu')
    instance_of_MLPInterface.model.add_frontend(frontend_layer)
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5)
    instance_of_MLPInterface.test_generator.batch_count = 0 #reset batch generator
    X,y = next(instance_of_MLPInterface.test_generator)

    pred, scores = instance_of_MLPInterface.run_on_batch(X, transform_input=True, return_raw_output=False)
    assert pred.shape == (5,)
    assert np.array_equal(pred, np.array([0,1,1,0,0]))

    assert scores.shape == (5,)
    np.testing.assert_array_almost_equal(scores, np.array([0.5122282, 0.9999663, 0.9999663, 0.5122282, 0.5122282], dtype=scores.dtype), decimal=7)

def test_run_on_batch_raw_output(instance_of_MLPInterface):
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5)
    X,_ = next(instance_of_MLPInterface.test_generator)

    scores = instance_of_MLPInterface.run_on_batch(X, transform_input=True, return_raw_output=True)

    assert scores.shape == (5,2)
    assert np.allclose(scores, np.array([[5.0865251e-01, 4.9134743e-01],
                                            [7.2768415e-12, 1.0000000e+00],
                                            [7.2768415e-12, 1.0000000e+00],
                                            [5.0865251e-01, 4.9134743e-01],
                                            [5.0865251e-01, 4.9134743e-01]], dtype=scores.dtype), atol=1e-08)

def test_run_on_instance(instance_of_MLPInterface):
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5)
    instance_of_MLPInterface.test_generator.batch_count = 0 #reset batch generator
    X,_ = next(instance_of_MLPInterface.test_generator)
    x = X[0:1]
    
    pred, scores = instance_of_MLPInterface.run_on_instance(x, return_raw_output=False)

    assert pred.shape == (1,)
    assert pred == 0
    assert scores.shape == (1,)
    np.testing.assert_array_almost_equal(scores, np.array([0.5086525], dtype=scores.dtype), decimal=7)

def test_run_on_instance_raw_output(instance_of_MLPInterface):
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5)
    instance_of_MLPInterface.test_generator.batch_count = 0 #reset batch generator
    X,_ = next(instance_of_MLPInterface.test_generator)
    x = X[0]
    
    scores = instance_of_MLPInterface.run_on_instance(x, return_raw_output=True)
    assert scores.shape == (1,2)
    np.testing.assert_array_almost_equal(scores, np.array([[0.5086525 , 0.49134743]], dtype=scores.dtype), decimal=7)
     
def test_run_on_test_generator(instance_of_MLPInterface):
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5)
    pred, scores = instance_of_MLPInterface.run_on_test_generator(compute_val_metrics=True, return_raw_output=False)

    assert pred.shape == (20,)
    assert np.array_equal(pred, np.array([0, 1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 1, 1, 1, 0, 1, 0, 1]))

    assert scores.shape == (20,)
    np.testing.assert_array_almost_equal(scores, np.array([0.5086525, 1.,1., 0.5086525, 0.5086525,
                                          0.5086525, 1., 0.5086525, 1.,0.5086525,
                                          0.5086525, 1., 0.5086525, 1., 1.,
                                          1., 0.5086525, 1., 0.5086525, 1.], dtype=scores.dtype), decimal=7)

def test_run_on_test_generator_metrics(instance_of_MLPInterface):
    """ Test that the run_on_test_generator method returns validation metrics """
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5)
    _, metrics = instance_of_MLPInterface.run_on_test_generator(compute_val_metrics=True, return_raw_output=False, return_metrics=True)
    
    assert metrics['CategoricalAccuracy'] == 1.0
    assert metrics['loss'].numpy() == pytest.approx(0.2456744, abs=0.00001)

def test_run_on_test_generator_raw_output(instance_of_MLPInterface):
    instance_of_MLPInterface.checkpoint_dir = os.path.join(path_to_tmp, "test_train_loop_checkpoints")
    instance_of_MLPInterface.train_loop(n_epochs=5)
    scores = instance_of_MLPInterface.run_on_test_generator(compute_val_metrics=True, return_raw_output=True)

    assert scores.shape == (4,5,2,)
    assert np.allclose(scores, np.array([[[5.0865251e-01, 4.9134743e-01],
                                            [7.2768415e-12, 1.0000000e+00],
                                            [7.2768415e-12, 1.0000000e+00],
                                            [5.0865251e-01, 4.9134743e-01],
                                            [5.0865251e-01, 4.9134743e-01]],

                                            [[5.0865251e-01, 4.9134743e-01],
                                            [7.2768415e-12, 1.0000000e+00],
                                            [5.0865251e-01, 4.9134743e-01],
                                            [7.2768415e-12, 1.0000000e+00],
                                            [5.0865251e-01, 4.9134743e-01]],

                                            [[5.0865251e-01, 4.9134743e-01],
                                            [7.2768415e-12, 1.0000000e+00],
                                            [5.0865251e-01, 4.9134743e-01],
                                            [7.2768415e-12, 1.0000000e+00],
                                            [7.2768415e-12, 1.0000000e+00]],

                                            [[7.2768415e-12, 1.0000000e+00],
                                            [5.0865251e-01, 4.9134743e-01],
                                            [7.2768415e-12, 1.0000000e+00],
                                            [5.0865251e-01, 4.9134743e-01],
                                            [7.2768415e-12, 1.0000000e+00]]], dtype=scores.dtype), atol=1e-08)



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Parsing module within the ketos library

    This module provides utilities to parse various string 
    structures.
"""
import os
import sys
import json
from pint import UnitRegistry
import importlib
import inspect

ureg = UnitRegistry()


""" Standard audio-representation parameters recognized by Ketos.
"""
audio_std_params = {'rate':                     {'type':float, 'unit':'Hz'},
                    'window':                   {'type':float, 'unit':'s'},
                    'step':                     {'type':float, 'unit':'s'},
                    'bins_per_oct':             {'type':int,   'unit':None},
                    'freq_min':                 {'type':float, 'unit':'Hz'},
                    'freq_max':                 {'type':float, 'unit':'Hz'},
                    'window_func':              {'type':str,   'unit':None},
                    'resample_method':          {'type':str,   'unit':None},
                    'duration':                 {'type':float, 'unit':'s'},
                    'normalize_wav':            {'type':bool,  'unit':None},
                    'transforms':               {'type':list,  'unit':None},
                    'waveform_transforms':      {'type':list,  'unit':None},
                    'num_chan':                 {'type':int,   'unit':None},
                    'filter_pad_samples':       {'type':int,   'unit':None},
                    'global_km_window_seconds': {'type':float, 'unit':'s'},
                    'local_km_window_seconds':  {'type':float, 'unit':'s'},
                    'filter_n':                 {'type':int,   'unit':None},
                    'filter_min_hz':            {'type':float, 'unit':'Hz'},
                    'decibel':                  {'type':bool,  'unit':None},
                    'input_shape':              {'type':list,  'unit':None}
                    }

def is_encoded(s):
    """ Check that the audio presentation has been encoded.

        More specifically, the method checks that items specified as having a 
        physical unit in `audio_std_params`, have string values.
    
        Args:
            s: dict
                Audio representation  

        Returns:
            : bool
                True, if the audio representation is encoded. False, otherwise.
    """
    s_dict = {'s': s} if 'type' in s.keys() else s
        
    for _,s in s_dict.items():     
        for key, value in s.items():
            if key in audio_std_params.keys() and \
                    audio_std_params[key]['unit'] != None and not isinstance(value, str):
                return False

    return True


def load_audio_representation(path, name=None, return_unparsed=False):
    """ Load audio representation from JSON file.

        By default the function attempts to parse the individual parameter 
        values, e.g., the value "20 kHz" will be returned as 20000 and the 
        value "11 ms" will be returned as 0.011. Use the `return_unparsed` 
        argument to change this behaviour.

        Args:
            path: str
                Path to json file
            name: str
                Heading of the relevant section of the json file. If None, 
                the function returns the entire content of the JSON file.
            return_unparsed: bool
                Do not parse the parameter values. Default is False.

        Returns:
            d: dict
                Audio representation 

        Example:
            >>> import json
            >>> import os
            >>> from ketos.data_handling.parsing import load_audio_representation
            >>> # create json file with spectrogram settings
            >>> json_str = '{"spectrogram": {"type": "MagSpectrogram", "rate": "20 kHz", "window": "0.1 s", "step": "0.025 s", "window_func": "hamming", "freq_min": "30Hz", "freq_max": "3000Hz"}}'
            >>> path = 'ketos/tests/assets/tmp/config.py'
            >>> file = open(path, 'w')
            >>> _ = file.write(json_str)
            >>> file.close()
            >>> # load settings back from json file
            >>> settings = load_audio_representation(path=path, name='spectrogram')
            >>> print(settings)
            {'type': <class 'ketos.audio.spectrogram.MagSpectrogram'>, 'rate': 20000.0, 'window': 0.1, 'step': 0.025, 'window_func': 'hamming', 'freq_min': 30, 'freq_max': 3000}
            >>> # clean up
            >>> os.remove(path)

            It is also possible to pass a custom audio representation class to this function. 
            In this case, include a key/value pair indicating the path to the module you are 
            loading the class from. For instance:

            >>> import json # doctest: +SKIP
            >>> import os # doctest: +SKIP
            >>> from ketos.data_handling.parsing import load_audio_representation # doctest: +SKIP
            >>> # create json file with spectrogram settings
            >>> json_str = '{"custom_representation": {"type": "Cepstrum", "module": "path/to/my/audio_representation.py", "any": "parameter", "for": "the", "custom": "representation"}}' # doctest: +SKIP
            >>> path = 'my/custom/config.py' # doctest: +SKIP
            >>> settings = load_audio_representation(path=path, name='custom_representation') # doctest: +SKIP
            >>> print(settings) # doctest: +SKIP
            {'type': <class 'audio_representation.Cepstrum'>, "module": "path/to/my/audio_representation.py", "any": "parameter", "for": "the", "custom": "representation"}
    """
    with open(path, 'r') as fil:
        data = json.load(fil)
        if name != None: 
            data = data[name]
        if not return_unparsed:
            data = parse_audio_representation(data)

    return data

def parse_audio_representation(audio_representations):
    """ Parse audio representation parameters.
    
        Args:
            audio_representation: dict
                Unparsed audio representation  

        Returns:
            audio_representation: dict
                Parsed audio representation
    """
    from ketos.audio import audio_representation_names_in_recipe
    # Determines if the input is a nested dictionary.    
    is_nested = isinstance(audio_representations, dict) and isinstance(list(audio_representations.values())[0], dict)

    if not is_nested:
        audio_representations = {0: audio_representations}

    for name,params in audio_representations.items():
        # check if audio representation type is a class included in ketos and return the class
        if params['type'] in audio_representation_names_in_recipe:
            audio_representations[name]['type'] = audio_representation_names_in_recipe[params['type']]
        else:
            try:
                # If not, try to load a custom module provided by the user
                # See docs https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
                module_name = os.path.basename(params['module']).split('.')[0]
                spec = importlib.util.spec_from_file_location(module_name, params['module'])
                module = importlib.util.module_from_spec(spec)
                sys.modules[module_name] = module
                spec.loader.exec_module(module)

                # Now we load the class
                audio_representation_class = getattr(module, params['type'])
                audio_representations[name]['type'] = audio_representation_class
            except KeyError as ex:   
                raise Exception(f'The audio representation "{audio_representations[name]["type"]}" is not included with ketos. However, it is possible to use a custom audio representation. Consult the documentation in "data_handling.parsing.load_audio_representation" for examples.') from None

        for key,value in params.items():
            audio_representations[name][key] = parse_parameter(name=key, value=value)

    if not is_nested:
        audio_representations = audio_representations[0]

    return audio_representations

def parse_parameter(name, value):
    """ Parse the parameter value according to the type and unit specified 
        in the `audio_std_params` dictionary. For example, if name='window' 
        and value='22.1 ms', the function returns the float 0.0221.

        If the parameter is not found in the `audio_std_params` dictionary, 
        the function returns the input value unmodified.

        Args:
            name: str
                Name of the parameter to be parsed
            value: str
                Value of the parameter to be parsed
            
        Returns:
            parsed_value: str, int, float, bool, or list 
                Parsed value

        Example:
            >>> from ketos.data_handling.parsing import parse_parameter
            >>> print(parse_parameter(name='step', value='23 ms'))
            0.023
    """
    Q = ureg.Quantity
    parsed_value = value

    if name in audio_std_params.keys():
        param = audio_std_params[name]
        typ  = param['type'] 
        unit = param['unit']

        if unit is not None and Q(value).check(unit): 
            parsed_value = Q(value).m_as(unit)

        if typ in ['int', int]:
            parsed_value = int(parsed_value)

        elif unit in ['float', float]:
            parsed_value = float(parsed_value)

        elif typ in ['str', str] and value is not None:
            parsed_value = str(parsed_value)

        elif typ in ['bool', bool]:
            parsed_value = (parsed_value.lower() == "true")

        elif typ in [list]:
            # convert specific transform arguments from str to tuple
            if name == 'transforms':
                for tr in parsed_value:
                    if tr['name'] == 'adjust_range':
                        s = tr['range'][1:-1]
                        tr['range'] = tuple(map(int, s.split(',')))

                    elif tr['name'] == 'resize' and 'shape' in tr.keys():
                        v = tr['shape']
                        assert isinstance(v, (list, str)), "shape argument of resize transform must be "\
                            f"of type 'list' or 'str' whereas a '{type(v)}' was provided"
                        if isinstance(v, list):
                            tr['shape'] = tuple(v)
                        elif isinstance(v, str):
                            s = v[1:-1]
                            tr['shape'] = tuple(map(int, s.split(',')))

    return parsed_value

def encode_audio_representation(s):
    """ Encode audio representation.

        Every parameter listed in the `audio_std_params` dictionary 
        with a unit is encoded as a str.
    
        Args:
            s: dict
                Input audio representation  

        Returns:
            s: dict
                Encoded audio representation
    """
    s_dict = {'s': s} if 'type' in s.keys() else s     
    for i,s in s_dict.items():         
        for key,value in s.items():
            s_dict[i][key] = encode_parameter(name=key, value=value)

    s = s_dict['s'] if 's' in s_dict.keys() else s_dict
    return s

def encode_parameter(name, value):
    """ Encode paramater as a string with an SI unit, according to the 
        unit specified in the `audio_std_params` dictionary. For example, 
        if name='window' and value=4.22, the function returns the str '4.22 s'.

        If the parameter is not found in the `audio_std_params` dictionary, 
        the function returns the input value unmodified, unless the parameter 
        is a tuple in which case it is converted to a string.
    
        Args:
            name: str
                Name of the parameter to be encoded
            value: str
                Value of the parameter to be encoded
            
        Returns:
            encoded_value: str or type of input value 
                Encoded value

        Example:
            >>> from ketos.data_handling.parsing import encode_parameter
            >>> print(encode_parameter(name='step', value=0.037))
            0.037 s
    """
    encoded_value = value
    if name in audio_std_params.keys():
        param = audio_std_params[name]
        unit = param['unit']
        if unit is not None:
            encoded_value = f'{value} {unit}'
        typ = param['type']
        if typ == bool:
            encoded_value = str(value).lower()

    elif isinstance(value, tuple):
        encoded_value = ','.join([str(x) for x in value])
        encoded_value = '(' + encoded_value + ')'
    
    elif name == 'type':
        if inspect.isclass(value):
            encoded_value = value.__name__
    return encoded_value

def str2bool(v):
    """ Convert most common answers to yes/no questions to boolean

    Args:
        v : str
            Answer 
    
    Returns:
        res : bool
            Answer converted to boolean 
    """
    res = v.lower() in ("yes", "YES", "Yes", "true", "True", "TRUE", "on", "ON", "t", "T", "1")
    return res




# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'neural_networks.dev_utils.export' module within the ketos library
"""
import pytest
import os
import numpy as np
import shutil
from zipfile import ZipFile
from pathlib import Path
from ketos.neural_networks import load_model_file
from ketos.audio import load_audio_representation_from_file
from ketos.data_handling.parsing import load_audio_representation
import ketos.neural_networks.dev_utils.export as exp

current_dir = Path(__file__).parent
path_to_assets = (current_dir / '..' / 'assets').resolve() # resolve converts the path to an absolute path
# path_to_tmp = path_to_assets / 'tmp'

def test_export_to_ketos_protobuf(tmpdir):
    """Test export resnet to ketos-protobuf format"""
    # model, audio_repr = load_model_file(model_path, tmp_path, load_audio_repr=True)
    model_path = path_to_assets / 'narw_resnet.kt'
    model = load_model_file(model_path)
    audio_repr = load_audio_representation_from_file(model_path)[0]['spectrogram']
    output_path = tmpdir / 'narw1.ktpb'

    exp.export_to_ketos_protobuf(model=model, output_name=output_path, audio_repr=audio_repr, 
        input_shape=(94,129), backward_compat=False)
    
    #check that file exists:
    assert (output_path).is_file()
    
    # check that file has correct content:
    with ZipFile(output_path, 'r') as zip:
        zip.extractall(path=tmpdir)
    
    assert (tmpdir / 'model').is_dir()
    assert (tmpdir / 'recipe.json').is_file()
    assert (tmpdir / 'audio_repr.json').is_file()

    # load audio representation
    audio_repr = load_audio_representation((tmpdir / 'audio_repr.json'), return_unparsed=True)
    assert 'spectrogram' in audio_repr

    # check that duration was written to file
    a = audio_repr['spectrogram']
    assert 'duration' in a
    assert a['duration'] == "3.008 s"

    # check that input_shape was written to file
    assert 'input_shape' in a
    assert a['input_shape'] == [1,94,129,1]

    # shutil.rmtree(tmp_path) #clean up


def test_export_to_ketos_protobuf_backward_compat(tmpdir):
    """Test export resnet to ketos-protobuf format backward compatible"""
    model_path = (path_to_assets / 'narw_resnet.kt')
    model, audio_repr = load_model_file(model_path, tmpdir, load_audio_repr=True)
    output_path = (tmpdir / 'narw1.ktpb')

    exp.export_to_ketos_protobuf(model=model, output_name=output_path, audio_repr=audio_repr[0], 
        input_shape=(1,94,129,1))
    
    assert (output_path).is_file()


def test_export_to_ketos_protobuf_infer_shape(tmpdir):
    """Test export resnet to ketos-protobuf format can infer shape"""
    model_path = path_to_assets / 'narw_resnet.kt'
    model, audio_repr = load_model_file(model_path, tmpdir, load_audio_repr=True)
    output_path = tmpdir / 'narw1.ktpb'

    with pytest.raises(AssertionError):
        exp.export_to_ketos_protobuf(model=model, output_name=output_path, audio_repr=audio_repr[0])

    exp.export_to_ketos_protobuf(model=model, output_name=output_path, audio_repr=audio_repr[0], duration=3.0)

    # check that file exists:
    assert (output_path).is_file()

    # check that shape was inferred correctly
    with ZipFile(output_path, 'r') as zip:
        zip.extractall(path=tmpdir)
    audio_repr = load_audio_representation(tmpdir / 'audio_repr.json')
    assert audio_repr['spectrogram']['input_shape'] == [1,94,129,1]


def test_export_to_ketos_protobuf_audio_repr(tmpdir):
    """Test export resnet to ketos-protobuf format using audio representation file path"""
    model_path = path_to_assets / 'narw_resnet.kt'
    audio_repr_path = path_to_assets / 'audio_repr.json'
    
    model = load_model_file(model_path, tmpdir)
    output_path = tmpdir / 'narw2.ktpb'

    exp.export_to_ketos_protobuf(model=model, output_name=output_path, audio_repr=audio_repr_path, 
        overwrite=True, duration=3.0)

    assert (output_path).is_file


def test_export_to_protobuf(tmpdir):
    """Test export resnet to protobuf format"""
    model_path = path_to_assets / 'narw_resnet.kt'
    model = load_model_file(model_path, tmpdir)
    input_spec = np.ones(shape=(94,129))
    model.run_on_instance(input_spec)
    output_path = tmpdir / 'model.pb'

    exp.export_to_protobuf(model=model, output_name=str(output_path))

    assert output_path.is_dir()


def test_export_to_ketos(tmpdir):
    """Test export resnet to ketos format"""
    model_path = path_to_assets / 'narw_resnet.kt'
    model = load_model_file(model_path)
    audio_repr = load_audio_representation_from_file(model_path)[0]['spectrogram']

    # check that we can save model
    output_path = tmpdir / 'narw3.kt'
    exp.export_to_ketos(model=model, output_name=output_path)
    assert os.path.isfile(output_path)

    # check that we can save specific checkpoint
    output_path = tmpdir / 'narw4.kt'
    exp.export_to_ketos(model=model, output_name=output_path, checkpoint_name="cp-0030.ckpt")
    assert os.path.isfile(output_path)

    # check that audio representation is saved correctly
    output_path = tmpdir / 'narw5.kt'
    exp.export_to_ketos(model=model, output_name=output_path, audio_repr=audio_repr)
    assert os.path.isfile(output_path)

    with ZipFile(output_path, 'r') as zip:
        zip.extractall(path=tmpdir)
    audio_repr = load_audio_representation(tmpdir / 'audio_repr.json')
    assert audio_repr['window'] == 0.256

    # check that we can save extra files to the .kt archive
    output_path = tmpdir / 'narw6.kt'
    exp.export_to_ketos(model=model, output_name=output_path, extra=(path_to_assets / "annot_001.csv"))
    assert os.path.isfile(output_path)

    with ZipFile(output_path, 'r') as zip:
        zip.extractall(path=tmpdir)
    assert (tmpdir / "annot_001.csv").exists()
    

def test_ketos_save_custom_module(tmpdir):
    """ Check if custom files are save correctly to the custom folder
        when using the custom_module parameter
    """
    model_path = (path_to_assets / 'narw_resnet.kt')
    model = load_model_file(model_path)

    custom_module = tmpdir / 'custom_module'
    os.mkdir(custom_module)
    (custom_module / 'another_file.py').touch()
    (custom_module / 'nn_architecture.py').touch()

    output_path = tmpdir / 'narw.kt'
    exp.export_to_ketos(model=model, output_name=output_path, custom_module=custom_module)
    with ZipFile(output_path, 'r') as archive:
        assert 'custom/nn_architecture.py' in archive.namelist()
        assert 'custom/another_file.py' in archive.namelist()

def test_get_export_function():
    """Test that we can detect the appropriate export function"""
    assert exp.get_export_function('/test/out.pb') == exp.export_to_protobuf
    assert exp.get_export_function('/test/out.kt') == exp.export_to_ketos
    assert exp.get_export_function('/test/out.ktpb') == exp.export_to_ketos_protobuf
    assert exp.get_export_function('/test/out.aaa') == exp.export_to_ketos
